"""
Example showing how to use the easing functions for position.

See:
https://easings.net/
...for a great guide on the theory behind how easings can work.

See example 2 for how to use easings for angles.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.easing_example_1
"""
import arcade
from arcade import easing
from arcade.types import Color

SPRITE_SCALING = 0.5

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Easing Example"

BACKGROUND_COLOR = "#F5D167"
TEXT_COLOR = "#4B1DF2"
BALL_COLOR = "#42B5EB"
LINE_COLOR = "#45E6D0"
LINE_WIDTH = 3

X_START = 40
X_END = 1200
Y_INTERVAL = 60
BALL_RADIUS = 13
TIME = 3.0


class EasingCircle(arcade.SpriteCircle):
    """ Player class """

    def __init__(self, radius, color):
        """ Set up the player """

        # Call the parent init
        super().__init__(radius, color)

        self.easing_x_data = None
        self.easing_y_data = None

    def update(self, delta_time: float = 1 / 60):
        if self.easing_x_data is not None:
            done, self.center_x = easing.ease_update(self.easing_x_data, delta_time)
            if done:
                x = X_START
                if self.center_x < WINDOW_WIDTH / 2:
                    x = X_END
                ex, ey = easing.ease_position(self.position,
                                              (x, self.center_y),
                                              rate=180,
                                              ease_function=self.easing_x_data.ease_function)
                self.easing_x_data = ex

        if self.easing_y_data is not None:
            done, self.center_y = easing.ease_update(self.easing_y_data, delta_time)
            if done:
                self.easing_y_data = None


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        """ Initializer """

        # Call the parent class initializer
        super().__init__()

        # Set the background color
        self.background_color = Color.from_hex_string(BACKGROUND_COLOR)

        self.ball_list = None
        self.text_list = []
        self.lines = None

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Sprite lists
        self.ball_list = arcade.SpriteList()
        self.lines = arcade.shape_list.ShapeElementList()

        def create_ball(ball_y, ease_function):
            ball = EasingCircle(BALL_RADIUS, Color.from_hex_string(BALL_COLOR))
            ball.position = X_START, ball_y
            p1 = ball.position
            p2 = (X_END, ball_y)
            ex, ey = easing.ease_position(p1, p2, time=TIME, ease_function=ease_function)
            ball.ease_function = ease_function
            ball.easing_x_data = ex
            ball.easing_y_data = ey
            return ball

        def create_line(line_y):
            line = arcade.shape_list.create_line(
                X_START, line_y - BALL_RADIUS - LINE_WIDTH,
                X_END, line_y - BALL_RADIUS,
                line_color, line_width=LINE_WIDTH,
            )
            return line

        def create_text(text_string):
            text = arcade.Text(
                text_string,
                x=X_START,
                y=y - BALL_RADIUS,
                color=text_color,
                font_size=24,
            )
            return text

        def add_item(item_y, ease_function, text):
            ball = create_ball(item_y, ease_function)
            self.ball_list.append(ball)
            text = create_text(text)
            self.text_list.append(text)
            line = create_line(item_y)
            self.lines.append(line)

        text_color = Color.from_hex_string(TEXT_COLOR)
        line_color = Color.from_hex_string(LINE_COLOR)

        y = Y_INTERVAL
        add_item(y, easing.linear, "Linear")

        y += Y_INTERVAL
        add_item(y, easing.ease_out, "Ease out")

        y += Y_INTERVAL
        add_item(y, easing.ease_in, "Ease in")

        y += Y_INTERVAL
        add_item(y, easing.smoothstep, "Smoothstep")

        y += Y_INTERVAL
        add_item(y, easing.ease_in_out, "Ease in/out")

        y += Y_INTERVAL
        add_item(y, easing.ease_out_elastic, "Ease out elastic")

        y += Y_INTERVAL
        add_item(y, easing.ease_in_back, "Ease in back")

        y += Y_INTERVAL
        add_item(y, easing.ease_out_back, "Ease out back")

        y += Y_INTERVAL
        add_item(y, easing.ease_in_sin, "Ease in sin")

        y += Y_INTERVAL
        add_item(y, easing.ease_out_sin, "Ease out sin")

        y += Y_INTERVAL
        add_item(y, easing.ease_in_out_sin, "Ease in out sin")

    def on_draw(self):
        """ Render the screen. """

        # This command has to happen before we start drawing
        self.clear()

        self.lines.draw()

        # Draw all the sprites.
        self.ball_list.draw()

        for text in self.text_list:
            text.draw()

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Call update on all sprites (The sprites don't do much in this
        # example though.)
        self.ball_list.update(delta_time)


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Example showing how to do transitions between views.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.transitions
"""
import arcade

WIDTH = 1280
HEIGHT = 720
FADE_RATE = 5


class FadingView(arcade.View):
    def __init__(self):
        super().__init__()
        self.fade_out = None
        self.fade_in = 255

    def update_fade(self, next_view=None):
        if self.fade_out is not None:
            self.fade_out += FADE_RATE
            if self.fade_out is not None and self.fade_out > 255 and next_view is not None:
                game_view = next_view()
                game_view.setup()
                self.window.show_view(game_view)

        if self.fade_in is not None:
            self.fade_in -= FADE_RATE
            if self.fade_in <= 0:
                self.fade_in = None

    def draw_fading(self):
        if self.fade_out is not None:
            arcade.draw_rect_filled(
                arcade.XYWH(
                    self.window.width / 2,
                    self.window.height / 2,
                    self.window.width,
                    self.window.height,
                ),
                color=(0, 0, 0, self.fade_out),
            )

        if self.fade_in is not None:
            arcade.draw_rect_filled(
                arcade.XYWH(
                    self.window.width / 2,
                    self.window.height / 2,
                    self.window.width,
                    self.window.height,
                ),
                color=(0, 0, 0, self.fade_in),
            )


class MenuView(FadingView):
    """ Class that manages the 'menu' view. """

    def on_update(self, dt):
        self.update_fade(next_view=GameView)

    def on_show_view(self):
        """ Called when switching to this view"""
        self.window.background_color = arcade.color.WHITE

    def on_draw(self):
        """ Draw the menu """
        self.clear()
        arcade.draw_text("Menu Screen - press space to advance", WIDTH / 2, HEIGHT / 2,
                         arcade.color.BLACK, font_size=30, anchor_x="center")
        self.draw_fading()

    def on_key_press(self, key, _modifiers):
        """ Handle key presses. In this case, we'll just count a 'space' as
        game over and advance to the game over view. """
        if self.fade_out is None and key == arcade.key.SPACE:
            self.fade_out = 0

    def setup(self):
        """ This should set up your game and get it ready to play """
        # Replace 'pass' with the code to set up your game
        pass


class GameView(FadingView):
    """ Manage the 'game' view for our program. """

    def setup(self):
        """ This should set up your game and get it ready to play """
        # Replace 'pass' with the code to set up your game
        pass

    def on_update(self, dt):
        self.update_fade(next_view=GameOverView)

    def on_show_view(self):
        """ Called when switching to this view"""
        self.window.background_color = arcade.color.ORANGE_PEEL

    def on_draw(self):
        """ Draw everything for the game. """
        self.clear()
        arcade.draw_text("Game - press SPACE to advance", WIDTH / 2, HEIGHT / 2,
                         arcade.color.BLACK, font_size=30, anchor_x="center")
        self.draw_fading()

    def on_key_press(self, key, _modifiers):
        """ Handle key presses. In this case, we'll just count a 'space' as
        game over and advance to the game overview. """
        if key == arcade.key.SPACE:
            self.fade_out = 0


class GameOverView(FadingView):
    """ Class to manage the game overview """
    def on_update(self, dt):
        self.update_fade(next_view=MenuView)

    def on_show_view(self):
        """ Called when switching to this view"""
        self.background_color = arcade.color.BLACK

    def on_draw(self):
        """ Draw the game overview """
        self.clear()
        arcade.draw_text("Game Over - press SPACE to advance", WIDTH / 2, HEIGHT / 2,
                         arcade.color.WHITE, 30, anchor_x="center")
        self.draw_fading()

    def on_key_press(self, key, _modifiers):
        """ If user hits escape, go back to the main menu view """
        if key == arcade.key.SPACE:
            self.fade_out = 0

    def setup(self):
        """ This should set up your game and get it ready to play """
        # Replace 'pass' with the code to set up your game
        pass


def main():
    """ Startup """
    window = arcade.Window(WIDTH, HEIGHT, "Different Views Minimal Example")
    menu_view = MenuView()
    window.show_view(menu_view)
    arcade.run()


if __name__ == "__main__":
    main()



"""
ScreenShakeController2D:
    Provides an easy way to cause a camera to shake.
"""

from math import exp, floor, log, pi, sin
from random import randint, uniform

from arcade.camera.data_types import CameraData
from arcade.math import quaternion_rotation

__all__ = ("ScreenShake2D",)


class ScreenShake2D:
    """
    Offsets the camera position in a random direction repeatedly over
    a set length of time to create a screen shake effect.

    The amplitude of the screen-shaking grows based on two functions.
    The first is a simple sin wave whose frequency is adjustable.
    This is multiplied with a pair of equations which go from 0-1 smoothly.
    the equation rises using a inverse exponential equation, before decreasing
    using a modified smooth-step sigmoid.

    Args:
        camera_data:
            The CameraData PoD that the controller modifies. Should not be
            changed once initialized.
        max_amplitude:
            The largest possible world space offset.
        falloff_time:
            The length of time in seconds it takes the shaking to reach 0 after
            reaching the maximum. Can be set to a negative number to disable falloff.
        acceleration_duration:
            The length of time in seconds it takes the shaking to reach max
            amplitude. Can be set to 0.0 to start at max amplitude.
        shake_frequency:
            The number of peaks per second. Avoid making it a multiple of half
            the target frame-rate. (e.g. at 60 fps avoid 30, 60, 90, 120, etc.)
        shake_angle:
            The angle around which the shake will focus. Defaults to 0 (horizontal)
            This is actuall 0 or 180 degress as the shake angle is mirrored.
        shake_range:
            The range in degrees from the shake angle that the shake may fall within.
            defaults to 90 degrees. This gives a range of +/- 90 degrees which covers the
            full circle because the shake angle is mirrored 180 degrees.
    """

    def __init__(
        self,
        camera_data: CameraData,
        *,
        max_amplitude: float = 1.0,
        falloff_time: float = 1.0,
        acceleration_duration: float = 1.0,
        shake_frequency: float = 15.0,
        shake_angle: float = 90.0,
        shake_range: float = 90.0,
    ):
        self._data: CameraData = camera_data

        self.max_amplitude: float = max_amplitude
        """The largest possible world space offset."""

        self.falloff_duration: float = falloff_time
        """
        The length of time in seconds it takes the shaking to reach 0
        after reaching the maximum.
        """

        self.shake_frequency: float = shake_frequency
        """
        The number of peaks per second. Avoid making it a multiple of
        half the target frame-rate.
        """

        self._acceleration_duration: float = acceleration_duration

        self.shake_angle: float = shake_angle
        self.shake_range: float = shake_range

        self._shaking: bool = False
        self._length_shaking: float = 0.0

        self._current_dir: float = 0.0
        self._last_vector: tuple[float, float, float] = (0.0, 0.0, 0.0)
        self._last_update_time: float = 0.0

    @property
    def shaking(self) -> bool:
        """Read only property to check if the controller is currently shaking the camera."""
        return self._shaking

    @property
    def duration(self) -> float:
        """
        The length of the screen shake in seconds.

        If falloff is disabled (by setting falloff_duration to a negative number) only returns the
        acceleration duration.

        Setting the duration to a negative number disables falloff.
        While the falloff is disabled setting duration will only set the acceleration.
        Otherwise, scales both the acceleration and falloff time to match new duration.
        """
        if self.falloff_duration < 0.0:
            return self._acceleration_duration
        return self._acceleration_duration + self.falloff_duration

    @duration.setter
    def duration(self, _duration: float) -> None:
        if _duration <= 0.0:
            self.falloff_duration = -1.0

        elif self.falloff_duration < 0.0:
            self._acceleration_duration = _duration
            return

        else:
            ratio = _duration / self.duration
            self._acceleration_duration = ratio * self._acceleration_duration
            self.falloff_duration = ratio * self.falloff_duration

    @property
    def current_amplitude(self) -> float:
        """Read only property which provides the current shake amplitude."""
        return self._calc_amplitude() * self.max_amplitude

    @property
    def acceleration_duration(self) -> float:
        """
        The length of time in seconds it takes for the shaking to reach max amplitude.

        Setting to a value less than zero causes the amplitude to start at max.
        """
        return self._acceleration_duration

    @acceleration_duration.setter
    def acceleration_duration(self, _duration: float) -> None:
        if _duration < 0.0:
            self._acceleration_duration = 0.0
        else:
            self._acceleration_duration = _duration

    @property
    def acceleration(self) -> float:
        """
        The inverse of acceleration time.

        setting to a value less than zero causes the amplitude to start at max.
        """
        if self._acceleration_duration <= 0.0:
            return 0.0
        return 1 / self._acceleration_duration

    @acceleration.setter
    def acceleration(self, _acceleration: float) -> None:
        if _acceleration <= 0.0:
            self._acceleration_duration = 0.0
        else:
            self._acceleration_duration = 1 / _acceleration

    @property
    def falloff(self) -> float:
        """
        The maximum gradient of the amplitude falloff,
        and is the gradient at the inflection point of the sigmoid equation.

        Is inversely proportional to the falloff duration by a factor of 15/8.
        """
        if self.falloff_duration < 0.0:
            return -1.0
        return (15 / 8) * (1 / self.falloff_duration)

    @falloff.setter
    def falloff(self, _falloff: float) -> None:
        if _falloff <= 0.0:
            self.falloff_duration = -1.0
        else:
            self.falloff_duration = (15 / 8) * (1 / _falloff)

    def _acceleration_amp(self, _t: float) -> float:
        """
        The equation for the growing half of the amplitude equation.
        It uses 1.0001 so that at _t = 1.0 the amplitude equals 1.0.

        Args:
            _t: The scaled time. Should be between 0.0 and 1.0
        """
        return 1.0001 - 1.0001 * exp(log(0.0001 / 1.0001) * _t)

    def _falloff_amp(self, _t: float) -> float:
        """
        The equation for the falloff half of the amplitude equation.
        It is based on the 'smootherstep' function.

        Args:
            _t: The scaled time. Should be between 0.0 and 1.0
        """
        return 1 - _t**3 * (_t * (_t * 6.0 - 15.0) + 10.0)

    def _calc_max_amp(self) -> float:
        """
        Determine the maximum amplitude by using either _acceleration_amp() or _falloff_amp().
        If falloff duration is less than 0.0 then the falloff never begins and
        """
        if self._length_shaking <= self._acceleration_duration:
            _t = self._length_shaking / self._acceleration_duration
            return self._acceleration_amp(_t)

        if self.falloff_duration < 0.0:
            return self.max_amplitude

        if self._length_shaking <= self.duration:
            _t = (self._length_shaking - self._acceleration_duration) / self.falloff_duration
            return self._falloff_amp(_t)

        return 0.0

    def _calc_amplitude(self) -> float:
        _max_amp = self._calc_max_amp()
        _sin_amp = sin(self.shake_frequency * 2.0 * pi * self._length_shaking)

        return _sin_amp * _max_amp

    def reset(self) -> None:
        """
        Reset the temporary shaking variables. WILL NOT STOP OR START SCREEN SHAKE.
        """
        self._current_dir = 0.0
        self._last_vector = (0.0, 0.0, 0.0)
        self._last_update_time = 0.0
        self._length_shaking = 0.0

    def start(self) -> None:
        """
        Start the screen-shake.
        """
        self.reset()
        self._shaking = True

    def stop(self) -> None:
        """
        Instantly stop the screen-shake.
        """
        self._data.position = (
            self._data.position[0] - self._last_vector[0],
            self._data.position[1] - self._last_vector[1],
            self._data.position[2] - self._last_vector[2],
        )

        self.reset()
        self._shaking = False

    def update(self, delta_time: float) -> None:
        """
        Update the time, and decide if the shaking should stop.
        Does not actually set the camera position.
        Should not be called more than once an update cycle.

        Args:
            delta_time:
                the length of time in seconds between update calls. Generally pass
                in the delta_time provided by the arcade.Window's on_update method.
        """
        if not self._shaking:
            return

        self._length_shaking += delta_time

        if self.falloff_duration > 0.0 and self._length_shaking >= self.duration:
            self.stop()

    def update_camera(self) -> None:
        """
        Update the position of the camera. Call this just before using the camera.
        because the controller is modifying the PoD directly it stores the last
        offset and resets the camera's position before adding the next offset.
        """
        if not self._shaking:
            return

        if (
            floor(self._last_update_time * 2 * self.shake_frequency)
            < floor(self._length_shaking * 2.0 * self.shake_frequency)
        ) or self._last_update_time == 0.0:
            # Start from the shake angle and then offset by the reflected shake range,
            # then randomly pick whether to mirror to the other side.
            self._current_dir = (
                self.shake_angle
                + uniform(-self.shake_range, self.shake_range)
                + (180 * randint(0, 1))
            )

        _amp = self._calc_amplitude() * self.max_amplitude
        _vec = quaternion_rotation(self._data.forward, self._data.up, self._current_dir)

        _last = self._last_vector
        _pos = self._data.position

        self._data.position = (
            _pos[0] - _last[0] + _vec[0] * _amp,
            _pos[1] - _last[1] + _vec[1] * _amp,
            _pos[2] - _last[2] + _vec[2] * _amp,
        )

        self._last_vector = (_vec[0] * _amp, _vec[1] * _amp, _vec[2] * _amp)
        self._last_update_time = self._length_shaking

    def readjust_camera(self) -> None:
        """
        Can be called after the camera has been used revert the screen_shake.
        While not strictly necessary it is highly advisable. If you are moving the
        camera using an animation or something similar the behavior can start to go
        awry if you do not readjust after the screen shake.
        """
        self._data.position = (
            self._data.position[0] - self._last_vector[0],
            self._data.position[1] - self._last_vector[1],
            self._data.position[2] - self._last_vector[2],
        )
        self._last_vector = (0.0, 0.0, 0.0)

    def __enter__(self):
        self.update_camera()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.readjust_camera()
        return False



"""
Use sprites to scroll around a large screen.

Simple program to show basic sprite usage.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.full_screen_example
"""

import arcade
from arcade.types import LRBT

SPRITE_SCALING = 0.5

WINDOW_WIDTH = 800
WINDOW_HEIGHT = 600
WINDOW_TITLE = "Full Screen Example"

# How many pixels to keep as a minimum margin between the character
# and the edge of the screen.
VIEWPORT_MARGIN = 40

MOVEMENT_SPEED = 5


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        """
        Initializer
        """
        # Open a window in full screen mode. Remove fullscreen=True if
        # you don't want to start this way.
        super().__init__()

        # This will get the size of the window, and set the viewport to match.
        # So if the window is 1000x1000, then so will our viewport. If
        # you want something different, then use those coordinates instead.
        self.background_color = arcade.color.AMAZON
        self.example_image = arcade.load_texture(":resources:images/tiles/boxCrate_double.png")

        # The camera used to update the viewport and projection on screen resize.
        self.camera = arcade.Camera2D(
            position=(0, 0),
            projection=LRBT(left=0, right=WINDOW_WIDTH, bottom=0, top=WINDOW_HEIGHT),
            viewport=self.window.rect
        )

    def on_draw(self):
        """
        Render the screen.
        """

        self.clear()
        with self.camera.activate():
            # Draw some boxes on the bottom so we can see how they change
            for count in range(20):
                x = count * 128
                y = count * 128
                width = 128
                height = 128
                arcade.draw_texture_rect(self.example_image, arcade.XYWH(x, y, width, height))

            arcade.draw_rect_outline(
                LRBT(0, WINDOW_WIDTH, 0, WINDOW_HEIGHT),
                color=arcade.color.WHITE,
                border_width=5,
            )

            # Draw text on the screen so the user has an idea of what is happening
            text_size = 18
            arcade.draw_text(
                "Press F to toggle between full screen and windowed mode, unstretched.",
                x=WINDOW_WIDTH // 2,
                y=WINDOW_HEIGHT // 2 - 20,
                color=arcade.color.WHITE,
                font_size=text_size,
                anchor_x="center",
            )
            arcade.draw_text(
                "Press S to toggle between full screen and windowed mode, stretched.",
                x=WINDOW_WIDTH // 2,
                y=WINDOW_HEIGHT // 2 + 20,
                color=arcade.color.WHITE,
                font_size=text_size,
                anchor_x="center",
            )

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed. """
        if key == arcade.key.F:
            # User hits f. Flip between full and not full screen.
            self.window.set_fullscreen(not self.window.fullscreen)

            # Get the window coordinates. Match viewport to window coordinates
            # so there is a one-to-one mapping.
            self.camera.viewport = self.window.rect
            self.camera.projection = arcade.LRBT(0.0, self.width, 0.0, self.height)

        if key == arcade.key.S:
            # User hits s. Flip between full and not full screen.
            self.window.set_fullscreen(not self.window.fullscreen)

            # Instead of a one-to-one mapping, stretch/squash window to match the
            # constants. This does NOT respect aspect ratio. You'd need to
            # do a bit of math for that.
            self.camera.projection = LRBT(
                left=0,
                right=WINDOW_WIDTH,
                bottom=0,
                top=WINDOW_HEIGHT,
            )
            self.camera.viewport = self.window.rect

        if key == arcade.key.ESCAPE:
            self.window.close()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on f
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Section Example 2:

In this Section example we develop a very basic Pong game playable by two
persons in the same computer (hot seat!).

Each Player is abstracted as a Section that consists of a space in the screen
where the player paddle can move.

Note:
    - Sections can live along with Views. Sections do not need to occupy 100%
      of the screen.
    - View methods can interact with Sections by storing a reference to each
      one.
    - How keyboard events can be redirected to each section depending on the
      pressed key automatically.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sections_demo_2
"""

import random

import arcade
from arcade.color import BLACK, BLUE, RED, BEAU_BLUE, GRAY
from arcade.key import W, S, UP, DOWN

PLAYER_SECTION_WIDTH = 100
PLAYER_PADDLE_SPEED = 10


class Player(arcade.Section):
    """
    A Section representing the space in the screen where the player
    paddle can move
    """

    def __init__(
        self, left: int, bottom: int, width: int, height: int, key_up: int, key_down: int, **kwargs
    ):
        super().__init__(
            left,
            bottom,
            width,
            height,
            accept_keyboard_keys={key_up, key_down},
            **kwargs,
        )

        # keys assigned to move the paddle
        self.key_up: int = key_up
        self.key_down: int = key_down

        # the player paddle
        self.paddle: arcade.SpriteSolidColor = arcade.SpriteSolidColor(30, 100, color=BLACK)

        # player score
        self.score: int = 0

    def setup(self):
        # reset the player paddle position to the middle of the screen
        self.paddle.position = self.left + 50, self.height / 2

    def on_update(self, delta_time: float):
        # update the paddle position
        self.paddle.update()

    def on_draw(self):
        # draw sections info and score
        if self.name == "Left":
            keys = "W and S"
            start_x = self.left + 5
        else:
            keys = "UP and DOWN"
            start_x = self.left - 290
        arcade.draw_text(
            f"Player {self.name} (move paddle with: {keys})",
            start_x,
            self.top - 20,
            BLUE,
            9,
        )
        arcade.draw_text(
            f"Score: {self.score}",
            self.left + 20,
            self.bottom + 20,
            BLUE,
        )

        # draw the paddle
        arcade.draw_sprite(self.paddle)

    def on_key_press(self, symbol: int, modifiers: int):
        # set the paddle direction and movement speed
        if symbol == self.key_up:
            self.paddle.change_y = PLAYER_PADDLE_SPEED
        else:
            self.paddle.change_y = -PLAYER_PADDLE_SPEED

    def on_key_release(self, _symbol: int, _modifiers: int):
        # stop moving the paddle
        self.paddle.stop()


class GameView(arcade.View):
    def __init__(self):
        super().__init__()

        # a sprite list that will hold each player paddle to
        # check for collisions
        self.paddles: arcade.SpriteList = arcade.SpriteList()

        # we store each Section
        self.left_player: Player = Player(
            0, 0, PLAYER_SECTION_WIDTH, self.window.height, key_up=W, key_down=S, name="Left"
        )
        self.right_player: Player = Player(
            self.window.width - PLAYER_SECTION_WIDTH,
            0,
            PLAYER_SECTION_WIDTH,
            self.window.height,
            key_up=UP,
            key_down=DOWN,
            name="Right",
        )

        # add the sections to the SectionManager so Sections start to work
        self.section_manager = arcade.SectionManager(self)
        self.section_manager.add_section(self.left_player)
        self.section_manager.add_section(self.right_player)

        # add each paddle to the sprite list
        self.paddles.append(self.left_player.paddle)
        self.paddles.append(self.right_player.paddle)

        # create the ball
        self.ball: arcade.SpriteCircle = arcade.SpriteCircle(20, RED)

    def setup(self):
        # set up a new game

        # set ball position in the middle
        self.ball.position = self.window.width / 2, self.window.height / 2

        # randomize ball direction and speed
        self.ball.change_x = random.choice([-3, -2, 3, 2])
        self.ball.change_y = random.choice([-3, -2, 3, 2])

        # setup player paddles
        self.left_player.setup()
        self.right_player.setup()

    def on_show_view(self) -> None:
        self.section_manager.enable()

    def on_hide_view(self) -> None:
        self.section_manager.disable()

    def on_update(self, delta_time: float):
        self.ball.update()  # update the ball

        # bounce the ball either at the top or at the bottom
        if self.ball.bottom <= 0:
            self.ball.change_y *= -1
        elif self.ball.top >= self.window.height:
            self.ball.change_y *= -1

        # check if the ball has collided with a paddle
        collided_paddle = self.ball.collides_with_list(self.paddles)
        if collided_paddle:
            # adjust ball coordinates to simplify the game
            if collided_paddle[0] is self.left_player.paddle:
                self.ball.left = self.left_player.paddle.right
            else:
                self.ball.right = self.right_player.paddle.left

            # bounce the ball from the paddle
            self.ball.change_x *= -1

        # check if the ball has exited the screen in either side and
        # end the game
        if self.ball.right <= 0:
            self.end_game(self.right_player)
        elif self.ball.left >= self.window.width:
            self.end_game(self.left_player)

    def end_game(self, winner: Player):
        """Called when one player wins"""
        winner.score += 1  # increment the winner score
        self.setup()  # prepare a new game

    def on_draw(self):
        self.clear(color=BEAU_BLUE)  # clear the screen

        arcade.draw_sprite(self.ball)  # draw the ball

        half_window_x = self.window.width / 2  # middle x

        # draw a line diving the screen in half
        arcade.draw_line(half_window_x, 0, half_window_x, self.window.height, GRAY, 2)


def main():
    # create the window
    window = arcade.Window(title="Two player simple Pong with Sections!")

    # create the custom View
    game = GameView()
    game.setup()

    # show the view
    window.show_view(game)

    # run arcade loop
    window.run()


if __name__ == "__main__":
    main()



"""
Example of Pymunk Physics Engine
Top-down

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.pymunk_demo_top_down
"""
import math
import random
import arcade
from arcade.pymunk_physics_engine import PymunkPhysicsEngine

WINDOW_TITLE = "PyMunk Top-Down"
SPRITE_SCALING_PLAYER = 0.5
MOVEMENT_SPEED = 5

SPRITE_IMAGE_SIZE = 128
SPRITE_SIZE = int(SPRITE_IMAGE_SIZE * SPRITE_SCALING_PLAYER)

WINDOW_WIDTH = SPRITE_SIZE * 15
WINDOW_HEIGHT = SPRITE_SIZE * 10

# Physics force used to move the player. Higher number, faster accelerating.
PLAYER_MOVE_FORCE = 4000
BULLET_MOVE_FORCE = 2500


class GameView(arcade.View):
    def __init__(self):
        """ Init """
        super().__init__()

        self.background_color = arcade.color.AMAZON

        self.player_list = None
        self.wall_list = None
        self.bullet_list = None
        self.rock_list = None
        self.gem_list = None
        self.player_sprite = None
        self.physics_engine: PymunkPhysicsEngine | None = None

        # Track the current state of what key is pressed
        self.left_pressed = False
        self.right_pressed = False
        self.up_pressed = False
        self.down_pressed = False

    def setup(self):
        """ Set up everything """
        # Create the sprite lists
        self.player_list = arcade.SpriteList()
        self.wall_list = arcade.SpriteList()
        self.bullet_list = arcade.SpriteList()
        self.rock_list = arcade.SpriteList()
        self.gem_list = arcade.SpriteList()

        # Set up the player
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=SPRITE_SCALING_PLAYER)
        self.player_sprite.center_x = 250
        self.player_sprite.center_y = 250
        self.player_list.append(self.player_sprite)

        # Set up the walls
        for x in range(0, WINDOW_WIDTH + 1, SPRITE_SIZE):
            wall = arcade.Sprite(":resources:images/tiles/grassCenter.png",
                                 scale=SPRITE_SCALING_PLAYER)
            wall.center_x = x
            wall.center_y = 0
            self.wall_list.append(wall)

            wall = arcade.Sprite(":resources:images/tiles/grassCenter.png",
                                 scale=SPRITE_SCALING_PLAYER)
            wall.center_x = x
            wall.center_y = WINDOW_HEIGHT
            self.wall_list.append(wall)

        # Set up the walls
        for y in range(SPRITE_SIZE, WINDOW_HEIGHT, SPRITE_SIZE):
            wall = arcade.Sprite(":resources:images/tiles/grassCenter.png",
                                 scale=SPRITE_SCALING_PLAYER)
            wall.center_x = 0
            wall.center_y = y
            self.wall_list.append(wall)

            wall = arcade.Sprite(":resources:images/tiles/grassCenter.png",
                                 scale=SPRITE_SCALING_PLAYER)
            wall.center_x = WINDOW_WIDTH
            wall.center_y = y
            self.wall_list.append(wall)

        # Add some movable rocks
        for x in range(SPRITE_SIZE * 2, SPRITE_SIZE * 13, SPRITE_SIZE):
            rock = random.randrange(4) + 1
            item = arcade.Sprite(f":resources:images/space_shooter/meteorGrey_big{rock}.png",
                                 scale=SPRITE_SCALING_PLAYER)
            item.center_x = x
            item.center_y = 400
            self.rock_list.append(item)

        # Add some movable coins
        for x in range(SPRITE_SIZE * 2, SPRITE_SIZE * 13, SPRITE_SIZE):
            items = [":resources:images/items/gemBlue.png",
                     ":resources:images/items/gemRed.png",
                     ":resources:images/items/coinGold.png",
                     ":resources:images/items/keyBlue.png"]
            item_name = random.choice(items)
            item = arcade.Sprite(item_name,
                                 scale=SPRITE_SCALING_PLAYER)
            item.center_x = x
            item.center_y = 300
            self.gem_list.append(item)

        # --- Pymunk Physics Engine Setup ---

        # The default damping for every object controls the percent of velocity
        # the object will keep each second. A value of 1.0 is no speed loss,
        # 0.9 is 10% per second, 0.1 is 90% per second.
        # For top-down games, this is basically the friction for moving objects.
        # For platformers with gravity, this should probably be set to 1.0.
        # Default value is 1.0 if not specified.
        damping = 0.7

        # Set the gravity. (0, 0) is good for outer space and top-down.
        gravity = (0, 0)

        # Create the physics engine
        self.physics_engine = PymunkPhysicsEngine(damping=damping,
                                                  gravity=gravity)

        def rock_hit_handler(sprite_a, sprite_b, arbiter, space, data):
            """ Called for bullet/rock collision """
            bullet_shape = arbiter.shapes[0]
            bullet_sprite = self.physics_engine.get_sprite_for_shape(bullet_shape)
            bullet_sprite.remove_from_sprite_lists()
            print("Rock")

        def wall_hit_handler(sprite_a, sprite_b, arbiter, space, data):
            """ Called for bullet/rock collision """
            bullet_shape = arbiter.shapes[0]
            bullet_sprite = self.physics_engine.get_sprite_for_shape(bullet_shape)
            bullet_sprite.remove_from_sprite_lists()
            print("Wall")

        self.physics_engine.add_collision_handler(
            "bullet",
            "rock",
            post_handler=rock_hit_handler,
        )
        self.physics_engine.add_collision_handler(
            "bullet",
            "wall",
            post_handler=wall_hit_handler,
        )

        # Add the player.
        # For the player, we set the damping to a lower value, which increases
        # the damping rate. This prevents the character from traveling too far
        # after the player lets off the movement keys.
        # Setting the moment to PymunkPhysicsEngine.MOMENT_INF prevents it from
        # rotating.
        # Friction normally goes between 0 (no friction) and 1.0 (high friction)
        # Friction is between two objects in contact. It is important to remember
        # in top-down games that friction moving along the 'floor' is controlled
        # by damping.
        self.physics_engine.add_sprite(self.player_sprite,
                                       friction=0.6,
                                       moment_of_inertia=PymunkPhysicsEngine.MOMENT_INF,
                                       damping=0.01,
                                       collision_type="player",
                                       max_velocity=400)

        # Create the walls.
        # By setting the body type to PymunkPhysicsEngine.STATIC the walls can't
        # move.
        # Movable objects that respond to forces are PymunkPhysicsEngine.DYNAMIC
        # PymunkPhysicsEngine.KINEMATIC objects will move, but are assumed to be
        # repositioned by code and don't respond to physics forces.
        # Dynamic is default.
        self.physics_engine.add_sprite_list(self.wall_list,
                                            friction=0.6,
                                            collision_type="wall",
                                            body_type=PymunkPhysicsEngine.STATIC)

        # Create some boxes to push around.
        # Mass controls, well, the mass of an object. Defaults to 1.
        self.physics_engine.add_sprite_list(self.rock_list,
                                            mass=2,
                                            friction=0.8,
                                            damping=0.1,
                                            collision_type="rock")
        # Create some boxes to push around.
        # Mass controls, well, the mass of an object. Defaults to 1.
        self.physics_engine.add_sprite_list(self.gem_list,
                                            mass=0.5,
                                            friction=0.8,
                                            damping=0.4,
                                            collision_type="rock")

    def on_mouse_press(self, x, y, button, modifiers):
        """ Called whenever the mouse button is clicked. """

        bullet = arcade.SpriteSolidColor(width=5, height=5, color=arcade.color.RED)
        self.bullet_list.append(bullet)

        # Position the bullet at the player's current location
        start_x = self.player_sprite.center_x
        start_y = self.player_sprite.center_y
        bullet.position = self.player_sprite.position

        # Get from the mouse the destination location for the bullet
        # IMPORTANT! If you have a scrolling screen, you will also need
        # to add in self.view_bottom and self.view_left.
        dest_x = x
        dest_y = y

        # Do math to calculate how to get the bullet to the destination.
        # Calculation the angle in radians between the start points
        # and end points. This is the angle the bullet will travel.
        x_diff = dest_x - start_x
        y_diff = dest_y - start_y
        angle = math.atan2(y_diff, x_diff)

        force = [math.cos(angle), math.sin(angle)]
        size = max(self.player_sprite.width, self.player_sprite.height) / 2

        bullet.center_x += size * force[0]
        bullet.center_y += size * force[1]

        self.physics_engine.add_sprite(bullet,
                                       mass=0.1,
                                       damping=1.0,
                                       friction=0.6,
                                       collision_type="bullet",
                                       elasticity=0.9)

        # Taking into account the angle, calculate our force.
        force[0] *= BULLET_MOVE_FORCE
        force[1] *= BULLET_MOVE_FORCE

        self.physics_engine.apply_force(bullet, force)

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed. """

        if key == arcade.key.UP:
            self.up_pressed = True
        elif key == arcade.key.DOWN:
            self.down_pressed = True
        elif key == arcade.key.LEFT:
            self.left_pressed = True
        elif key == arcade.key.RIGHT:
            self.right_pressed = True
        elif key == arcade.key.SPACE:
            bullet = arcade.SpriteSolidColor(9, 9, arcade.color.RED)
            bullet.position = self.player_sprite.position
            bullet.center_x += 30
            self.bullet_list.append(bullet)
            self.physics_engine.add_sprite(bullet,
                                           mass=0.2,
                                           damping=1.0,
                                           friction=0.6,
                                           collision_type="bullet")
            force = (3000, 0)
            self.physics_engine.apply_force(bullet, force)

    def on_key_release(self, key, modifiers):
        """Called when the user releases a key. """

        if key == arcade.key.UP:
            self.up_pressed = False
        elif key == arcade.key.DOWN:
            self.down_pressed = False
        elif key == arcade.key.LEFT:
            self.left_pressed = False
        elif key == arcade.key.RIGHT:
            self.right_pressed = False

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Calculate speed based on the keys pressed
        self.player_sprite.change_x = 0
        self.player_sprite.change_y = 0

        if self.up_pressed and not self.down_pressed:
            force = (0, PLAYER_MOVE_FORCE)
            self.physics_engine.apply_force(self.player_sprite, force)
        elif self.down_pressed and not self.up_pressed:
            force = (0, -PLAYER_MOVE_FORCE)
            self.physics_engine.apply_force(self.player_sprite, force)
        if self.left_pressed and not self.right_pressed:
            self.player_sprite.change_x = -MOVEMENT_SPEED
            force = (-PLAYER_MOVE_FORCE, 0)
            self.physics_engine.apply_force(self.player_sprite, force)
        elif self.right_pressed and not self.left_pressed:
            force = (PLAYER_MOVE_FORCE, 0)
            self.physics_engine.apply_force(self.player_sprite, force)

        # --- Move items in the physics engine
        self.physics_engine.step()

    def on_draw(self):
        """ Draw everything """
        self.clear()
        self.wall_list.draw()
        self.bullet_list.draw()
        self.rock_list.draw()
        self.gem_list.draw()
        self.player_list.draw()

def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Array Backed Grid Shown By Sprites

Show how to use a two-dimensional list/array to back the display of a
grid on-screen.

This version syncs the grid to the sprite list in one go using resync_grid_with_sprites.
This is faster than rebuilding a shape list every time the grid changes,
but we are still inspecting every single cell of the grid when it updates.
There are faster ways, but this works for smaller grid sizes.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.array_backed_grid_sprites_1
"""
import arcade

# Set how many rows and columns we will have
ROW_COUNT = 15
COLUMN_COUNT = 15

# This sets the WIDTH and HEIGHT of each grid location
WIDTH = 30
HEIGHT = 30

# This sets the margin between each cell
# and on the edges of the screen.
MARGIN = 5

# Do the math to figure out our screen dimensions
WINDOW_WIDTH = (WIDTH + MARGIN) * COLUMN_COUNT + MARGIN
WINDOW_HEIGHT = (HEIGHT + MARGIN) * ROW_COUNT + MARGIN
WINDOW_TITLE = "Array Backed Grid Example"


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        """
        Set up the application.
        """
        super().__init__()

        # Create a 2 dimensional array. A two dimensional
        # array is simply a list of lists.
        # This array can be altered later to contain 0 or 1
        # to show a white or green cell.
        #
        # A 4 x 4 grid would look like this
        #
        # grid = [
        #     [0, 0, 0, 0],
        #     [0, 0, 0, 0],
        #     [0, 0, 0, 0],
        #     [0, 0, 0, 0],
        # ]
        # We can quickly build a grid with python list comprehension
        # self.grid = [[0] * COLUMN_COUNT for _ in range(ROW_COUNT)]
        # Making the grid with loops:
        self.grid = []
        for row in range(ROW_COUNT):
            # Add an empty array that will hold each cell
            # in this row
            self.grid.append([])
            for column in range(COLUMN_COUNT):
                self.grid[row].append(0)  # Append a cell

        # Set the window's background color
        self.background_color = arcade.color.BLACK
        # Create a spritelist for batch drawing all the grid sprites
        self.grid_sprite_list = arcade.SpriteList()

        # Create a list of solid-color sprites to represent each grid location
        for row in range(ROW_COUNT):
            for column in range(COLUMN_COUNT):
                x = column * (WIDTH + MARGIN) + (WIDTH / 2 + MARGIN)
                y = row * (HEIGHT + MARGIN) + (HEIGHT / 2 + MARGIN)
                sprite = arcade.SpriteSolidColor(WIDTH, HEIGHT, color=arcade.color.WHITE)
                sprite.center_x = x
                sprite.center_y = y
                self.grid_sprite_list.append(sprite)

    def resync_grid_with_sprites(self):
        """
        Update the color of all the sprites to match
        the color/stats in the grid.

        We look at the values in each cell.
        If the cell contains 0 we assign a white color.
        If the cell contains 1 we assign a green color.
        """
        for row in range(ROW_COUNT):
            for column in range(COLUMN_COUNT):
                # We need to convert our two dimensional grid to our
                # one-dimensional sprite list. For example a 10x10 grid might have
                # row 2, column 8 mapped to location 28. (Zero-basing throws things
                # off, but you get the idea.)
                # ALTERNATIVELY you could set self.grid_sprite_list[pos].texture
                # to different textures to change the image instead of the color.
                pos = row * COLUMN_COUNT + column
                if self.grid[row][column] == 0:
                    self.grid_sprite_list[pos].color = arcade.color.WHITE
                else:
                    self.grid_sprite_list[pos].color = arcade.color.GREEN

    def on_draw(self):
        """
        Render the screen.
        """
        # We should always start by clearing the window pixels
        self.clear()

        # Batch draw all the sprites
        self.grid_sprite_list.draw()

    def on_mouse_press(self, x, y, button, modifiers):
        """
        Called when the user presses a mouse button.
        """

        # Convert the clicked mouse position into grid coordinates
        column = int(x // (WIDTH + MARGIN))
        row = int(y // (HEIGHT + MARGIN))

        print(f"Click coordinates: ({x}, {y}). Grid coordinates: ({row}, {column})")

        # Make sure we are on-grid. It is possible to click in the upper right
        # corner in the margin and go to a grid location that doesn't exist
        if row >= ROW_COUNT or column >= COLUMN_COUNT:
            # Simply return from this method since nothing needs updating
            return

        # Flip the location between 1 and 0.
        if self.grid[row][column] == 0:
            self.grid[row][column] = 1
        else:
            self.grid[row][column] = 0

        # Update the sprite colors to match the new grid
        self.resync_grid_with_sprites()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from __future__ import annotations

from collections.abc import Generator
from contextlib import contextmanager
from math import radians, tan
from typing import TYPE_CHECKING

from pyglet.math import Mat4, Vec2, Vec3
from typing_extensions import Self

from arcade.camera.data_types import DEFAULT_FAR, CameraData, PerspectiveProjectionData, Projector
from arcade.camera.projection_functions import (
    generate_perspective_matrix,
    generate_view_matrix,
    project_perspective,
    unproject_perspective,
)
from arcade.types import LBWH, Point, Rect
from arcade.window_commands import get_window

if TYPE_CHECKING:
    from arcade import Window


__all__ = ("PerspectiveProjector",)


class PerspectiveProjector(Projector):
    """
    The simplest from of a perspective camera.

    .. warning:: Near cutoffs for perspective projection must be greater than zero.

                 This prevents division by zero errors since perspective involves
                 dividing by distance.

    Using ViewData and PerspectiveProjectionData PoDs (Pack of Data)
    it generates the correct projection and view matrices. It also
    provides methods and a context manager for using the matrices in
    glsl shaders.

    This class provides no methods for manipulating the PoDs.

    The current implementation will recreate the view and
    projection matrices every time the camera is used.
    If used every frame or multiple times per frame this may
    be inefficient. If you suspect this is causing slowdowns
    profile before optimizing with a dirty value check.

    Initialize a Projector which produces a perspective projection matrix using
    a CameraData and PerspectiveProjectionData PoDs.

    Args:
        window:
            The window to bind the camera to. Defaults to the currently active camera.
        view:
            The CameraData PoD. contains the viewport, position, up, forward, and zoom.
        projection:
            The PerspectiveProjectionData PoD. contains the field of view, aspect ratio,
            and then near and far planes.
    """

    def __init__(
        self,
        *,
        window: Window | None = None,
        view: CameraData | None = None,
        projection: PerspectiveProjectionData | None = None,
        viewport: Rect | None = None,
        scissor: Rect | None = None,
    ):
        self._window: Window = window or get_window()

        self.viewport: Rect = viewport or LBWH(0, 0, self._window.width, self._window.height)
        self.scissor: Rect | None = scissor

        self._view = view or CameraData(  # Viewport
            (self._window.width / 2, self._window.height / 2, 0),  # Position
            (0.0, 1.0, 0.0),  # Up
            (0.0, 0.0, -1.0),  # Forward
            1.0,  # Zoom
        )

        self._projection = projection or PerspectiveProjectionData(
            self._window.width / self._window.height,  # Aspect
            60,  # Field of View,
            0.01,
            DEFAULT_FAR,  # near, # far
        )

    @property
    def view(self) -> CameraData:
        """Get the internal :py:class:`~arcade.camera.data_types.CameraData`.

        This is a read-only property.
        The CameraData. Is a read only property.
        """
        return self._view

    @property
    def projection(self) -> PerspectiveProjectionData:
        """Get the :py:class:`~arcade.camera.data_types.PerspectiveProjectionData`.

        This is a read-only property.
        """
        return self._projection

    def generate_projection_matrix(self) -> Mat4:
        """Generates a projection matrix.

        This is an alias of
        :py:meth:`generate_perspective_matrix`.
        """
        return generate_perspective_matrix(self._projection, self._view.zoom)

    def generate_view_matrix(self) -> Mat4:
        """Generates a view matrix.

        This is an alias of=
        :py:meth:`generate_view_matrix`.
        """
        return generate_view_matrix(self._view)

    @contextmanager
    def activate(self) -> Generator[Self, None, None]:
        """Set this camera as the current one, then undo it after.

        This method is a :external:ref:`context manager <context-managers>`
        you can use inside ``with`` blocks. Using it this way guarantees
        that the old camera and its settings will be restored, even if an
        exception occurs:

        .. code-block:: python

           # Despite an Exception, the previous camera and its settings
           # will be restored at the end of the with block below:
           with projector_instance.activate():
                sprite_list.draw()
                _ = 1 / 0  # Guaranteed ZeroDivisionError

        """
        previous_projector = self._window.current_camera
        try:
            self.use()
            yield self
        finally:
            previous_projector.use()

    def use(self) -> None:
        """Set the active camera to this object and apply other config.

        This includes the following steps:

        #. Set the window's current camera to this one
        #. Generate appropriate view and projection matrices
        #. Set the GL context's viewport and scissorbox values
        #. Apply the relevant matrices to Arcade's
           :py:class:`~arcade.Window` object
        """

        self._window.current_camera = self

        _projection = generate_perspective_matrix(self._projection, self._view.zoom)
        _view = generate_view_matrix(self._view)

        self._window.ctx.viewport = self.viewport.lbwh_int
        self._window.ctx.scissor = None if not self.scissor else self.scissor.lbwh_int
        self._window.projection = _projection
        self._window.view = _view

    def project(self, world_coordinate: Point) -> Vec2:
        """Convert world coordinates to pixel screen coordinates.

        If a 2D :py:class:`~pyglet.math.Vec2` is provided instead of a 3D
        :py:class:`~pyglet.math.Vec3`, then one will be calculated to the best of the
        method's ability.

        Args:
            world_coordinate:
                A :py:class:`~pyglet.math.Vec2` or :py:class:`~pyglet.math.Vec3` as world
                coordinates.

        Returns:
            A 2D screen pixel coordinate.
        """
        x, y, *_z = world_coordinate
        z = (
            (
                0.5
                * self.viewport.height
                / tan(radians(0.5 * self._projection.fov / self._view.zoom))
            )
            if not _z
            else _z[0]
        )

        _projection = generate_perspective_matrix(self._projection, self._view.zoom)
        _view = generate_view_matrix(self._view)

        pos = project_perspective(Vec3(x, y, z), self.viewport.lbwh_int, _view, _projection)

        return pos

    # TODO: update args
    def unproject(self, screen_coordinate: Point) -> Vec3:
        """Convert a pixel coordinate into world space.

        This reverses the effects of :py:meth:`.project`.

        Args:
            screen_coordinate:
                A 2D position in pixels from the bottom left of the screen.
                This should ALWAYS be in the range of 0.0 - screen size.

        Returns: A 3D vector in world space.

        """
        x, y, *_z = screen_coordinate
        z = (
            (
                0.5
                * self.viewport.height
                / tan(radians(0.5 * self._projection.fov / self._view.zoom))
            )
            if not _z
            else _z[0]
        )

        _projection = generate_perspective_matrix(self._projection, self._view.zoom)
        _view = generate_view_matrix(self._view)

        pos = unproject_perspective(Vec3(x, y, z), self.viewport.lbwh_int, _view, _projection)
        return pos



"""
This module contains commands for basic graphics drawing commands.
(Drawing primitives.)

Many of these commands are slow, because they load everything to the
graphics card each time a shape is drawn. For faster drawing, see the
Buffered Draw Commands.
"""

from .arc import draw_arc_filled, draw_arc_outline
from .parabola import draw_parabola_filled, draw_parabola_outline
from .circle import (
    draw_circle_filled,
    draw_circle_outline,
    draw_ellipse_filled,
    draw_ellipse_outline,
)
from .line import draw_line_strip, draw_line, draw_lines
from .point import draw_point, draw_points
from .polygon import draw_polygon_filled, draw_polygon_outline
from .triangle import draw_triangle_filled, draw_triangle_outline
from .rect import (
    draw_lrbt_rectangle_outline,
    draw_lbwh_rectangle_outline,
    draw_rect_outline,
    draw_lrbt_rectangle_filled,
    draw_lbwh_rectangle_filled,
    draw_rect_filled,
    draw_texture_rect,
    draw_sprite,
    draw_sprite_rect,
)
from .helpers import get_points_for_thick_line

__all__ = [
    # arc
    "draw_arc_filled",
    "draw_arc_outline",
    # parabola
    "draw_parabola_filled",
    "draw_parabola_outline",
    # circle
    "draw_circle_filled",
    "draw_circle_outline",
    "draw_ellipse_filled",
    "draw_ellipse_outline",
    # line
    "draw_line_strip",
    "draw_line",
    "draw_lines",
    # point
    "draw_point",
    "draw_points",
    # polygon
    "draw_polygon_filled",
    "draw_polygon_outline",
    # triangle
    "draw_triangle_filled",
    "draw_triangle_outline",
    # rectangle
    "draw_lrbt_rectangle_outline",
    "draw_lbwh_rectangle_outline",
    "draw_rect_outline",
    "draw_lrbt_rectangle_filled",
    "draw_lbwh_rectangle_filled",
    "draw_rect_filled",
    "draw_texture_rect",
    "draw_sprite",
    "draw_sprite_rect",
    # helpers
    "get_points_for_thick_line",
]



"""
Sprite Move With Walls

Simple program to show basic sprite usage.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_move_walls
"""

import arcade

SPRITE_SCALING = 0.5

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprite Move with Walls Example"

MOVEMENT_SPEED = 5


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        """
        Initializer
        """
        super().__init__()

        # Sprite lists
        self.coin_list = None
        self.wall_list = None
        self.player_list = None

        # Set up the player
        self.player_sprite = None
        self.physics_engine = None

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Sprite lists
        self.player_list = arcade.SpriteList()
        self.wall_list = arcade.SpriteList()

        # Set up the player
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=SPRITE_SCALING,
        )
        self.player_sprite.center_x = 50
        self.player_sprite.center_y = 64
        self.player_list.append(self.player_sprite)

        # -- Set up the walls
        # Create a row of boxes
        for x in range(173, 650, 64):
            wall = arcade.Sprite(":resources:images/tiles/boxCrate_double.png",
                                 scale=SPRITE_SCALING)
            wall.center_x = x
            wall.center_y = 200
            self.wall_list.append(wall)

        # Create a column of boxes
        for y in range(273, 500, 64):
            wall = arcade.Sprite(":resources:images/tiles/boxCrate_double.png",
                                 scale=SPRITE_SCALING)
            wall.center_x = 465
            wall.center_y = y
            self.wall_list.append(wall)

        self.physics_engine = arcade.PhysicsEngineSimple(self.player_sprite,
                                                         self.wall_list)

        # Set the background color
        self.background_color = arcade.color.AMAZON

    def on_draw(self):
        """
        Render the screen.
        """

        # This command has to happen before we start drawing
        self.clear()

        # Draw all the sprites.
        self.wall_list.draw()
        self.player_list.draw()

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed. """

        if key == arcade.key.UP:
            self.player_sprite.change_y = MOVEMENT_SPEED
        elif key == arcade.key.DOWN:
            self.player_sprite.change_y = -MOVEMENT_SPEED
        elif key == arcade.key.LEFT:
            self.player_sprite.change_x = -MOVEMENT_SPEED
        elif key == arcade.key.RIGHT:
            self.player_sprite.change_x = MOVEMENT_SPEED

    def on_key_release(self, key, modifiers):
        """Called when the user releases a key. """

        if key == arcade.key.UP or key == arcade.key.DOWN:
            self.player_sprite.change_y = 0
        elif key == arcade.key.LEFT or key == arcade.key.RIGHT:
            self.player_sprite.change_x = 0

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Call update on all sprites (The sprites don't do much in this
        # example though.)
        self.physics_engine.update()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Simple experimental profiler. This api is not stable.
"""

import cProfile
import pstats
from contextlib import contextmanager
from io import StringIO


class Profiler:
    """
    A profiler that can measure the execution time within a section
    of your code.

    Examples::

        # Context manager
        profiler = Profiler()
        with profiler.enabled():
            # Code you want to measure
        profiler.print_stats()

        # Enable and disable manually
        profiler = Profiler()
        profiler.enable()
        # Code you want to measure
        profiler.disable()
        profiler.print_stats()

        The same profiler instance can be enabled multiple times
        to accumulate data.

    Args:
        sort_by: function sort order
    """

    def __init__(self, sort_by="tottime"):
        self._sort_by = sort_by
        self._profiler = cProfile.Profile()

    @contextmanager
    def enabled(self):
        self.enable()
        try:
            yield
        finally:
            self.disable()

    def enable(self):
        """Start/enable the profiler"""
        self._profiler.enable()

    def disable(self):
        """Stop/disable the profiler"""
        self._profiler.disable()

    def print_stats(self):
        stats = pstats.Stats(self._profiler).sort_stats(self._sort_by)
        stats.print_stats()

    def get_stats_string(self):
        strdata = StringIO()
        pstats.Stats(self._profiler, stream=strdata).sort_stats(self._sort_by)
        return strdata.getvalue()



"""This example is a proof-of-concept for a scrollable area.

The example shows vertical and horizontal scroll areas with a list of buttons.

The current implementation lags a proper API, customizability, mouse support and documentation,
but shows how to use the current experimental feature.

If Arcade and Python are properly installed, you can run this example with:
python -m arcade.examples.gui.exp_scroll_area
"""

import arcade
from arcade.gui import UIAnchorLayout, UIBoxLayout, UIFlatButton, UIView
from arcade.gui.experimental import UIScrollArea
from arcade.gui.experimental.scroll_area import UIScrollBar


class MyView(UIView):
    def __init__(self):
        super().__init__()
        self.background_color = arcade.uicolor.BLUE_BELIZE_HOLE

        # create a layout with two columns
        root = self.add_widget(UIAnchorLayout())
        content_left = UIAnchorLayout(size_hint=(0.5, 1))
        root.add(content_left, anchor_x="left", anchor_y="center")
        content_right = UIAnchorLayout(size_hint=(0.5, 1))
        root.add(content_right, anchor_x="right", anchor_y="center")

        # create the list, which should be scrolled vertically
        vertical_list = UIBoxLayout(size_hint=(1, 0), space_between=1)
        for i in range(100):
            button = UIFlatButton(height=30, size_hint=(1, None), text=f"Button {i}")
            vertical_list.add(button)

        # the scroll area and the scrollbar are added to a box layout
        # so they are next to each other, this also reduces complexity for the layout
        # implementation
        v_scroll_area = UIBoxLayout(vertical=False, size_hint=(0.8, 0.8))
        content_left.add(v_scroll_area, anchor_x="center", anchor_y="center")

        scroll_layout = v_scroll_area.add(UIScrollArea(size_hint=(1, 1)))
        scroll_layout.with_border(color=arcade.uicolor.WHITE_CLOUDS)
        scroll_layout.add(vertical_list)

        v_scroll_area.add(UIScrollBar(scroll_layout))

        # create the list, which should be scrolled vertically
        horizontal_list = UIBoxLayout(size_hint=(0, 1), space_between=1, vertical=False)
        for i in range(100):
            button = UIFlatButton(width=50, size_hint=(None, 1), text=f"B{i}")
            horizontal_list.add(button)

        # same as above, but for horizontal scrolling
        h_scroll_area = UIBoxLayout(vertical=True, size_hint=(0.8, 0.8))
        content_right.add(h_scroll_area, anchor_x="center", anchor_y="center")

        h_scroll_layout = h_scroll_area.add(UIScrollArea(size_hint=(1, 1)))
        h_scroll_layout.with_border(color=arcade.uicolor.WHITE_CLOUDS)
        h_scroll_layout.add(horizontal_list)

        h_scroll_area.add(UIScrollBar(h_scroll_layout, vertical=False))

    def on_key_press(self, symbol: int, modifiers: int) -> bool | None:
        if symbol == arcade.key.ESCAPE:
            arcade.close_window()
            return True

        return False


def main():
    window = arcade.Window(title="GUI Example: UIScrollLayout")
    window.show_view(MyView())
    window.run()


if __name__ == "__main__":
    main()



"""
This animation example shows how perform a radar sweep animation.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.radar_sweep
"""
import arcade
import math

# Set up the constants
WINDOW_WIDTH = 800
WINDOW_HEIGHT = 600
WINDOW_TITLE = "Radar Sweep Example"

# These constants control the particulars about the radar
CENTER_X = WINDOW_WIDTH // 2
CENTER_Y = WINDOW_HEIGHT // 2
RADIANS_PER_FRAME = 0.02
SWEEP_LENGTH = 250


class Radar:
    def __init__(self):
        self.angle = 0

    def update(self, delta_time=0):
        # Move the angle of the sweep.
        self.angle += RADIANS_PER_FRAME * delta_time

    def draw(self):
        """ Use this function to draw everything to the screen. """

        # Calculate the end point of our radar sweep. Using math.
        x = SWEEP_LENGTH * math.sin(self.angle) + CENTER_X
        y = SWEEP_LENGTH * math.cos(self.angle) + CENTER_Y

        # Draw the radar line
        arcade.draw_line(CENTER_X, CENTER_Y, x, y, arcade.color.OLIVE, 4)

        # Draw the outline of the radar
        arcade.draw_circle_outline(CENTER_X,
                                   CENTER_Y,
                                   SWEEP_LENGTH,
                                   arcade.color.DARK_GREEN,
                                   border_width=10,
                                   num_segments=60)


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        super().__init__()

        # Create our rectangle
        self.radar = Radar()

        # Set background color
        self.background_color = arcade.color.BLACK

    def on_update(self, delta_time):
        # Move the rectangle
        self.radar.update(delta_time * 60)

    def on_draw(self):
        """Draw the screen"""
        # Clear screen
        self.clear()
        # Draw the rectangle
        self.radar.draw()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Dual-stick Shooter Example

A dual-analog stick controller is the preferred method of input. If a controller is
not present, the game will fail back to use keyboard controls (WASD to move, arrows to shoot)

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.dual_stick_shooter
"""
import math
import pprint
import random
from typing import cast

import arcade

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Dual-stick Shooter Example"
MOVEMENT_SPEED = 4
BULLET_SPEED = 10
BULLET_COOLDOWN_TICKS = 10
ENEMY_SPAWN_INTERVAL = 1
ENEMY_SPEED = 1
STICK_DEADZONE = 0.05
# An angle of "0" means "right", but the player's texture is oriented in the "up" direction.
# So an offset is needed.
ROTATE_OFFSET = -90


def dump_obj(obj):
    for key in sorted(vars(obj)):
        val = getattr(obj, key)
        print(f"{key:30} = {val} ({type(val).__name__})")


def dump_controller(controller):
    print(f"========== {controller}")
    print(f"Left X        {controller.leftx}")
    print(f"Left Y        {controller.lefty}")
    print(f"Left Trigger  {controller.lefttrigger}")
    print(f"Right X       {controller.rightx}")
    print(f"Right Y       {controller.righty}")
    print(f"Right Trigger {controller.righttrigger}")
    print("========== Extra controller")
    dump_obj(controller)
    print("========== Extra controller.device")
    dump_obj(controller.device)
    print("========== pprint controller")
    pprint.pprint(controller)
    print("========== pprint controller.device")
    pprint.pprint(controller.device)


def dump_controller_state(ticks, controller):
    # print("{:5.2f} {:5.2f} {:>20} {:5}_".format(1.234567, -8.2757272903, "hello", str(True)))
    fmt_str = "{:6d} "
    num_fmts = ["{:5.2f}"] * 6
    fmt_str += " ".join(num_fmts)
    print(fmt_str.format(ticks,
                         controller.leftx,
                         controller.lefty,
                         controller.lefttrigger,
                         controller.rightx,
                         controller.righty,
                         controller.righttrigger,
                         ))


def get_stick_position(x, y):
    """Given position of stick axes, return (x, y, angle_in_degrees).
    If movement is not outside of deadzone, return (None, None, None)"""
    if x > STICK_DEADZONE or x < -STICK_DEADZONE or y > STICK_DEADZONE or y < -STICK_DEADZONE:
        rad = math.atan2(y, x)
        angle = math.degrees(rad)
        return x, y, angle
    return None, None, None


class Player(arcade.sprite.Sprite):
    def __init__(self, filename):
        super().__init__(
            filename,
            scale=0.4,
            center_x=WINDOW_WIDTH / 2,
            center_y=WINDOW_HEIGHT / 2,
        )
        self.shoot_up_pressed = False
        self.shoot_down_pressed = False
        self.shoot_left_pressed = False
        self.shoot_right_pressed = False
        self.start_pressed = False


class Enemy(arcade.sprite.Sprite):
    def __init__(self, x, y):
        super().__init__(
            ':resources:images/pinball/bumper.png',
            scale=0.5,
            center_x=x,
            center_y=y,
        )

    def follow_sprite(self, player_sprite):
        """
        This function will move the current sprite towards whatever
        other sprite is specified as a parameter.

        We use the 'min' function here to get the sprite to line up with
        the target sprite, and not jump around if the sprite is not off
        an exact multiple of ENEMY_SPEED.
        """

        if self.center_y < player_sprite.center_y:
            self.center_y += min(ENEMY_SPEED, player_sprite.center_y - self.center_y)
        elif self.center_y > player_sprite.center_y:
            self.center_y -= min(ENEMY_SPEED, self.center_y - player_sprite.center_y)

        if self.center_x < player_sprite.center_x:
            self.center_x += min(ENEMY_SPEED, player_sprite.center_x - self.center_x)
        elif self.center_x > player_sprite.center_x:
            self.center_x -= min(ENEMY_SPEED, self.center_x - player_sprite.center_x)


class GameView(arcade.View):
    def __init__(self):
        super().__init__()
        self.game_over = False
        self.score = 0
        self.tick = 0
        self.bullet_cooldown = 0
        self.player = Player(":resources:images/space_shooter/playerShip2_orange.png")
        self.bullet_list = arcade.SpriteList()
        self.enemy_list = arcade.SpriteList()
        self.controller_manager = arcade.ControllerManager()
        self.controller = None

        self.background_color = arcade.color.DARK_MIDNIGHT_BLUE
        controllers = self.controller_manager.get_controllers()
        for controller in controllers:
            dump_controller(controller)
        if controllers:
            self.connect_controller(controllers[0])
            arcade.window_commands.schedule(self.debug_controller_state, 0.1)

        arcade.window_commands.schedule(self.spawn_enemy, ENEMY_SPAWN_INTERVAL)

        @self.controller_manager.event
        def on_connect(controller):
            if not self.controller:
                self.connect_controller(controller)

        @self.controller_manager.event
        def on_disconnect(controller):
            if self.controller == controller:
                controller.close()
                self.controller = None

    def setup(self):
        self.game_over = False
        self.score = 0
        self.tick = 0
        self.bullet_cooldown = 0
        self.bullet_list = arcade.SpriteList()
        self.enemy_list = arcade.SpriteList()
        self.player.center_x = WINDOW_WIDTH / 2
        self.player.center_y = WINDOW_HEIGHT / 2

    def connect_controller(self, controller):
        self.controller = controller
        self.controller.open()

    def debug_controller_state(self, _delta_time):
        if self.controller:
            dump_controller_state(self.tick, self.controller)

    def spawn_enemy(self, _elapsed):
        if self.game_over:
            return
        x = random.randint(0, WINDOW_WIDTH)
        y = random.randint(0, WINDOW_HEIGHT)
        self.enemy_list.append(Enemy(x, y))

    def on_update(self, delta_time):
        self.tick += 1
        if self.game_over:
            if self.controller:
                if self.controller.start:
                    self.setup()
                    return

            if self.player.start_pressed:
                self.setup()
                return

            return

        self.bullet_cooldown += 1

        for enemy in self.enemy_list:
            cast(Enemy, enemy).follow_sprite(self.player)

        if self.controller:
            # Controller input - movement
            move_x, move_y, move_angle = get_stick_position(
                self.controller.leftx, self.controller.lefty
            )
            if move_angle:
                self.player.change_x = move_x * MOVEMENT_SPEED
                self.player.change_y = move_y * MOVEMENT_SPEED
                self.player.angle = move_angle + ROTATE_OFFSET
            else:
                self.player.change_x = 0
                self.player.change_y = 0

            # Controller input - shooting
            shoot_x, shoot_y, shoot_angle = get_stick_position(
                self.controller.rightx, self.controller.righty
            )
            if shoot_angle:
                self.spawn_bullet(shoot_angle)
        else:
            # Keyboard input - shooting
            if self.player.shoot_right_pressed and self.player.shoot_up_pressed:
                self.spawn_bullet(0 + 45)
            elif self.player.shoot_up_pressed and self.player.shoot_left_pressed:
                self.spawn_bullet(90 + 45)
            elif self.player.shoot_left_pressed and self.player.shoot_down_pressed:
                self.spawn_bullet(180 + 45)
            elif self.player.shoot_down_pressed and self.player.shoot_right_pressed:
                self.spawn_bullet(270 + 45)
            elif self.player.shoot_right_pressed:
                self.spawn_bullet(0)
            elif self.player.shoot_up_pressed:
                self.spawn_bullet(90)
            elif self.player.shoot_left_pressed:
                self.spawn_bullet(180)
            elif self.player.shoot_down_pressed:
                self.spawn_bullet(270)

        self.enemy_list.update()
        self.player.update()
        self.bullet_list.update()
        ship_death_hit_list = arcade.check_for_collision_with_list(self.player,
                                                                   self.enemy_list)
        if len(ship_death_hit_list) > 0:
            self.game_over = True
        for bullet in self.bullet_list:
            bullet_killed = False
            enemy_shot_list = arcade.check_for_collision_with_list(bullet,
                                                                   self.enemy_list)
            # Loop through each colliding sprite, remove it, and add to the score.
            for enemy in enemy_shot_list:
                enemy.remove_from_sprite_lists()
                bullet.remove_from_sprite_lists()
                bullet_killed = True
                self.score += 1
            if bullet_killed:
                continue

    def on_key_press(self, key, modifiers):
        if key == arcade.key.W:
            self.player.change_y = MOVEMENT_SPEED
        elif key == arcade.key.A:
            self.player.change_x = -MOVEMENT_SPEED
        elif key == arcade.key.S:
            self.player.change_y = -MOVEMENT_SPEED
        elif key == arcade.key.D:
            self.player.change_x = MOVEMENT_SPEED
        elif key == arcade.key.RIGHT:
            self.player.shoot_right_pressed = True
        elif key == arcade.key.UP:
            self.player.shoot_up_pressed = True
        elif key == arcade.key.LEFT:
            self.player.shoot_left_pressed = True
        elif key == arcade.key.DOWN:
            self.player.shoot_down_pressed = True
        # close the window if the user hits the escape key
        elif key == arcade.key.ESCAPE:
            self.window.close()

        rad = math.atan2(self.player.change_y, self.player.change_x)
        self.player.angle = math.degrees(rad) + ROTATE_OFFSET

    def on_key_release(self, key, modifiers):
        if key == arcade.key.W:
            self.player.change_y = 0
        elif key == arcade.key.A:
            self.player.change_x = 0
        elif key == arcade.key.S:
            self.player.change_y = 0
        elif key == arcade.key.D:
            self.player.change_x = 0
        elif key == arcade.key.RIGHT:
            self.player.shoot_right_pressed = False
        elif key == arcade.key.UP:
            self.player.shoot_up_pressed = False
        elif key == arcade.key.LEFT:
            self.player.shoot_left_pressed = False
        elif key == arcade.key.DOWN:
            self.player.shoot_down_pressed = False

        rad = math.atan2(self.player.change_y, self.player.change_x)
        self.player.angle = math.degrees(rad) + ROTATE_OFFSET

    def spawn_bullet(self, angle_in_deg):
        # only allow bullet to spawn on an interval
        if self.bullet_cooldown < BULLET_COOLDOWN_TICKS:
            return
        self.bullet_cooldown = 0

        bullet = arcade.Sprite(":resources:images/space_shooter/laserBlue01.png", scale=0.75)

        # Position the bullet at the player's current location
        start_x = self.player.center_x
        start_y = self.player.center_y
        bullet.center_x = start_x
        bullet.center_y = start_y

        # angle the bullet visually
        bullet.angle = angle_in_deg
        angle_in_rad = math.radians(angle_in_deg)

        # set bullet's movement direction
        bullet.change_x = math.cos(angle_in_rad) * BULLET_SPEED
        bullet.change_y = math.sin(angle_in_rad) * BULLET_SPEED

        # Add the bullet to the appropriate lists
        self.bullet_list.append(bullet)

    def on_draw(self):
        # clear screen and start render process
        self.clear()

        # draw game items
        self.bullet_list.draw()
        self.enemy_list.draw()
        arcade.draw_sprite(self.player)

        # Put the score on the screen.
        output = f"Score: {self.score}"
        arcade.draw_text(output, 10, 20, arcade.color.WHITE, 14)

        # Game over message
        if self.game_over:
            arcade.draw_text("Game Over",
                             WINDOW_WIDTH / 2,
                             WINDOW_HEIGHT / 2,
                             arcade.color.WHITE, 100,
                             width=WINDOW_WIDTH,
                             align="center",
                             anchor_x="center",
                             anchor_y="center")



def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Controller API. This is basically just an alias to Pyglet's controller API.

For more info on this API, see
https://pyglet.readthedocs.io/en/latest/programming_guide/input.html#using-controllers
"""

import pyglet.input

__all__ = ["get_controllers", "ControllerManager"]


def get_controllers():
    """
    This returns a list of controllers, it is synonymous with calling
    ``pyglet.input.get_controllers()``
    """
    return pyglet.input.get_controllers()


class ControllerManager(pyglet.input.ControllerManager):
    """A ControllerManager provides an interface for handling connect/disconnect events.

    Please see Pyglet docs:
    https://pyglet.readthedocs.io/en/latest/programming_guide/input.html#controllermanager
    """

    pass



"""
Example showing how to draw localized text to the screen.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.text_loc_example_done
"""
import arcade
import gettext

# Try to auto-detect the user's language and translate to it
gettext.install('text_loc_example', localedir='text_loc_example_locale')

WINDOW_WIDTH = 500
WINDOW_HEIGHT = 500
WINDOW_TITLE = "Localizing Text Example"

class GameView(arcade.View):
    """
    Main application class.
    """
    def __init__(self):
        super().__init__()
        self.window.background_color = arcade.color.WHITE
        self.text = arcade.Text(
            gettext.gettext("Simple line of text in 12 point"),
            50.0, 450.0, arcade.color.BLACK, 12
        )

    def on_draw(self):
        """
        Render the screen.
        """
        # This command should happen before we start drawing. It will clear
        # the screen to the background color, and erase what we drew last frame.
        self.clear()

        # We draw a dot to make it clear how the text relates to the text
        arcade.draw_point(self.text.x, self.text.y, arcade.color.BLUE, 5)
        self.text.draw()

def main():
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)
    game = GameView()

    window.show_view(game)
    arcade.run()


if __name__ == "__main__":
    main()



from __future__ import annotations

import warnings
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Literal, TypeVar

from typing_extensions import override

from arcade.gui.property import bind, unbind
from arcade.gui.widgets import UILayout, UIWidget, _ChildEntry

__all__ = ["UILayout", "UIAnchorLayout", "UIBoxLayout", "UIGridLayout"]

W = TypeVar("W", bound="UIWidget")


class UIAnchorLayout(UILayout):
    """Places children based on anchor values.

    Defaults to ``size_hint = (1, 1)``.

    Supports the options ``size_hint``, ``size_hint_min``, and
    ``size_hint_max``. Children are allowed to overlap.

    Child are resized based on ``size_hint``. ``size_hint_min/max`` only take effect
    if a ``size_hint`` is set.

    Allowed keyword options for
    :py:meth:`~arcade.gui.UIAnchorLayout.add`:

    - ``anchor_x``: ``str`` = ``None``

      Horizontal anchor position for the layout. The class constant
      :py:attr:`~arcade.gui.UIAnchorLayout.default_anchor_x` is used as
      default.

    - ``anchor_y``: ``str`` = ``None``

      Vertical anchor position for the layout. The class constant
      :py:attr:`~arcade.gui.UIAnchorLayout.default_anchor_y` is used as
      default.

    - ``align_x``: ``float`` = 0

      Horizontal alignment for the layout.

    - ``align_y``: ``float`` = 0

      Vertical alignement for the layout.

    Args:
        x: ``x`` coordinate of the bottom left corner.
        y: ``y`` coordinate of the bottom left corner.
        width: Width of the layout.
        height: Height of the layout.
        children: Initial list of children. More can be added later.
        size_hint: Size hint for :py:class:`~arcade.gui.UILayout`
        size_hint_min: Minimum width and height in pixels.
        size_hint_max: Maximum width and height in pixels.
        **kwargs: Additional keyword arguments passed to UILayout.
    """

    default_anchor_x = "center"
    default_anchor_y = "center"
    _restrict_child_size = False
    """Whether to restrict the child size to the layout size.
    For scroll use cases this is not wanted, but for UITextWidgets it is.
    """

    def __init__(
        self,
        *,
        x: float = 0,
        y: float = 0,
        width: float = 1,
        height: float = 1,
        children: Iterable[UIWidget] = tuple(),
        size_hint=(1, 1),
        size_hint_min=None,
        size_hint_max=None,
        **kwargs,
    ):
        super().__init__(
            x=x,
            y=y,
            width=width,
            height=height,
            children=children,
            size_hint=size_hint,
            size_hint_min=size_hint_min,
            size_hint_max=size_hint_max,
            **kwargs,
        )

    def do_layout(self):
        """Executes the layout algorithm.

        Children are placed based on their anchor values."""
        for child, data in self._children:
            self._place_child(child, **data)

    def add(
        self,
        child: W,
        *,
        anchor_x: str | None = None,
        align_x: float = 0,
        anchor_y: str | None = None,
        align_y: float = 0,
        **kwargs,
    ) -> W:
        """Add a widget to the layout as a child. Added widgets will receive
        all user-interface events and be rendered.

        By default, the latest added widget will receive events first and will
        be rendered on top of others. The widgets will be automatically placed
        within this widget.

        Args:
            child: Specified child widget to add.
            anchor_x: Horizontal anchor. Valid options are ``left``,
                ``right``, and ``center``.
            align_x: Offset or padding for the horizontal anchor.
            anchor_y: Vertical anchor. Valid options are ``top``,
                ``center``, and ``bottom``.
            align_y: Offset or padding for the vertical anchor.

        Returns:
            Given child that was just added to the layout.
        """
        return super().add(
            child=child,
            anchor_x=anchor_x,
            align_x=align_x,
            anchor_y=anchor_y,
            align_y=align_y,
            **kwargs,
        )

    def _place_child(
        self,
        child: UIWidget,
        anchor_x: str | None = None,
        align_x: float = 0,
        anchor_y: str | None = None,
        align_y: float = 0,
    ):
        anchor_x = anchor_x or self.default_anchor_x
        anchor_y = anchor_y or self.default_anchor_y

        # Handle size_hints
        new_child_rect = child.rect

        sh_w, sh_h = child.size_hint or (None, None)
        shmn_w, shmn_h = child.size_hint_min or (None, None)
        shmx_w, shmx_h = child.size_hint_max or (None, None)

        if sh_w is not None:
            new_child_rect = new_child_rect.resize(width=self.content_width * sh_w)

            if shmn_w:
                new_child_rect = new_child_rect.min_size(width=shmn_w)
            if shmx_w:
                new_child_rect = new_child_rect.max_size(width=shmx_w)

            if self._restrict_child_size:
                new_child_rect = new_child_rect.max_size(width=self.content_width)

        if sh_h is not None:
            new_child_rect = new_child_rect.resize(height=self.content_height * sh_h)

            if shmn_h:
                new_child_rect = new_child_rect.min_size(height=shmn_h)
            if shmx_h:
                new_child_rect = new_child_rect.max_size(height=shmx_h)

            if self._restrict_child_size:
                new_child_rect = new_child_rect.max_size(height=self.content_height)

        # Calculate position
        content_rect = self.content_rect

        anchor_x = "center_x" if anchor_x == "center" else anchor_x
        child_anchor_x_value = getattr(new_child_rect, anchor_x)
        own_anchor_x_value = getattr(content_rect, anchor_x)
        diff_x = own_anchor_x_value + align_x - child_anchor_x_value

        anchor_y = "center_y" if anchor_y == "center" else anchor_y
        child_anchor_y_value = getattr(new_child_rect, anchor_y)
        own_anchor_y_value = getattr(content_rect, anchor_y)
        diff_y = own_anchor_y_value + align_y - child_anchor_y_value

        # Check if changes are required
        if diff_x or diff_y or child.rect != new_child_rect:
            child.rect = new_child_rect.move(diff_x, diff_y)


class UIBoxLayout(UILayout):
    """Place widgets next to each other. Depending on the
    :py:class:`~arcade.gui.UIBoxLayout.vertical` attribute, the widgets are
    placed top to bottom or left to right.

    .. hint::

        :py:class:`~arcade.gui.UIBoxLayout` does not adjust its
        own size if children are added. This requires a
        :py:class:`~arcade.gui.UIManager` or a
        :py:class:`~arcade.gui.UIAnchorLayout` as a parent.

        Or use :py:meth:`arcade.gui.UIBoxLayout.fit_content` to resize the layout. The
        bottom-left corner is used as the default anchor point.

    Supports the options: ``size_hint``, ``size_hint_min``, ``size_hint_max``. ``size_hint_min``
    is automatically updated based on the minimal required space by children.

    If a child widget provides a ``size_hint`` for a dimension, the child will
    be resized within the given range of ``size_hint_min`` and
    ``size_hint_max`` (unrestricted if not given). If the parameter
    ``vertical`` is True, any available space (``layout size - min_size`` of
    children) will be distributed to the child widgets based on their
    ``size_hint``.

    Args:
        x: ``x`` coordinate of the bottom left corner.
        y: ``y`` coordinate of the bottom left corner.
        vertical: Layout children vertical (True) or horizontal (False).
        align: Align children in orthogonal direction:: - ``x``:
            ``left``, ``center``, and ``right`` - ``y``: ``top``,
            ``center``, and ``bottom``
        children: Initial list of children. More can be added later.
        size_hint: Size hint for the :py:class:`~arcade.gui.UILayout` if
            the widget would like to grow. Defaults to ``0, 0`` ->
            minimal size to contain children.
        size_hint_max: Maximum width and height in pixels.
        space_between: Space in pixels between the children.
    """

    def __init__(
        self,
        *,
        x=0,
        y=0,
        width=1,
        height=1,
        vertical=True,
        align="center",
        children: Iterable[UIWidget] = tuple(),
        size_hint=(0, 0),
        size_hint_max=None,
        space_between=0,
        style=None,
        **kwargs,
    ):
        super().__init__(
            x=x,
            y=y,
            width=width,
            height=height,
            children=children,
            size_hint=size_hint,
            size_hint_max=size_hint_max,
            style=style,
            **kwargs,
        )

        self.align = align
        self.vertical = vertical
        self._space_between = space_between

        self._size_hint_requires_update = True

        bind(self, "_children", self._update_size_hints)
        bind(self, "_border_width", self._update_size_hints)

        bind(self, "_padding_left", self._update_size_hints)
        bind(self, "_padding_right", self._update_size_hints)
        bind(self, "_padding_top", self._update_size_hints)
        bind(self, "_padding_bottom", self._update_size_hints)

        self._update_size_hints()

    @override
    def add(self, child: W, **kwargs) -> W:
        """Add a widget to this layout

        Args:
            child: The widget to add to the layout.
        """
        # subscribe to child's changes, which might affect the own size hint
        bind(child, "_children", self._trigger_size_hint_update)
        bind(child, "rect", self._trigger_size_hint_update)
        bind(child, "size_hint", self._trigger_size_hint_update)
        bind(child, "size_hint_min", self._trigger_size_hint_update)
        bind(child, "size_hint_max", self._trigger_size_hint_update)

        return super().add(child, **kwargs)

    @override
    def remove(self, child: UIWidget):
        """Remove a child from the layout."""
        # unsubscribe from child's changes
        unbind(child, "_children", self._trigger_size_hint_update)
        unbind(child, "rect", self._trigger_size_hint_update)
        unbind(child, "size_hint", self._trigger_size_hint_update)
        unbind(child, "size_hint_min", self._trigger_size_hint_update)
        unbind(child, "size_hint_max", self._trigger_size_hint_update)

        return super().remove(child)

    def _trigger_size_hint_update(self):
        self._size_hint_requires_update = True

    def _update_size_hints(self):
        self._size_hint_requires_update = False

        required_space_between = max(0, len(self.children) - 1) * self._space_between
        min_child_sizes = [UILayout.min_size_of(child) for child in self.children]

        if len(self.children) == 0:
            width = 0
            height = 0
        elif self.vertical:
            width = max(size[0] for size in min_child_sizes)
            height_of_children = sum(size[1] for size in min_child_sizes)
            height = height_of_children + required_space_between
        else:
            width_of_children = sum(size[0] for size in min_child_sizes)
            width = width_of_children + required_space_between
            height = max(size[1] for size in min_child_sizes)

        base_width = self._padding_left + self._padding_right + 2 * self._border_width
        base_height = self._padding_top + self._padding_bottom + 2 * self._border_width
        self.size_hint_min = base_width + width, base_height + height

    def fit_content(self):
        """Resize the layout to fit the content.
        This will take the minimal required size into account."""
        self._update_size_hints()
        self.rect = self.rect.resize(self.size_hint_min[0], self.size_hint_min[1])

    def prepare_layout(self):
        """Updates the size hints if required."""
        super().prepare_layout()

        if self._size_hint_requires_update:
            self._update_size_hints()

    def do_layout(self):
        """Executes the layout algorithm.

        This method is called by the parent layout to place the children, after the rect was set.

        The layout algorithm will place the children based on the size hints next to each other.
        Depending on the vertical attribute, the children are placed top to bottom or left to right.
        """
        start_y = self.content_rect.top
        start_x = self.content_rect.left

        if not self.children:
            return

        # main axis
        constraints = [
            _C.from_widget_height(child) if self.vertical else _C.from_widget_width(child)
            for child, _ in self._children
        ]

        available_space = (
            self.content_height if self.vertical else self.content_width
        ) - self._space_between * (len(self.children) - 1)
        main_sizes = _box_axis_algorithm(constraints, available_space)

        # orthogonal axis
        constraints = [
            _C.from_widget_width(child) if self.vertical else _C.from_widget_height(child)
            for child, _ in self._children
        ]
        orthogonal_sizes = _box_orthogonal_algorithm(
            constraints, self.content_width if self.vertical else self.content_height
        )

        for (child, data), main_size, ortho_size in zip(
            self._children, main_sizes, orthogonal_sizes
        ):
            # apply calculated sizes, condition regarding existing size_hint
            # are already covered in calculation input
            new_rect = child.rect.resize(
                height=main_size if self.vertical else ortho_size,
                width=ortho_size if self.vertical else main_size,
            )
            new_rect = (
                new_rect.align_x(start_x + self.content_width / 2)
                if self.vertical
                else new_rect.align_y(start_y - self.content_height // 2)
            )

            # align on main and orthogonal axis and update start position
            if self.vertical:
                if self.align == "left":
                    new_rect = new_rect.align_left(start_x)
                elif self.align == "right":
                    new_rect = new_rect.align_right(start_x + self.content_width)
                else:
                    center_x = start_x + self.content_width // 2
                    new_rect = new_rect.align_x(center_x)
                new_rect = new_rect.align_top(start_y)
                start_y -= main_size + self._space_between
            else:
                if self.align == "top":
                    new_rect = new_rect.align_top(start_y)
                elif self.align == "bottom":
                    new_rect = new_rect.align_bottom(start_y - self.content_height)
                else:
                    center_y = start_y - self.content_height // 2
                    new_rect = new_rect.align_y(center_y)
                new_rect = new_rect.align_left(start_x)
                start_x += main_size + self._space_between

            # update child rect
            child.rect = new_rect

    def __str__(self):
        return f"UIBoxLayout(vertical={self.vertical})"


class UIGridLayout(UILayout):
    """Place widgets in a grid.

    Widgets can span multiple columns and rows.
    By default, the layout will only use the minimal required space (``size_hint = (0, 0)``).

    Widgets can provide a ``size_hint`` to request dynamic space relative to the layout size.
    A size_hint of ``(1, 1)`` will fill the available space, while ``(0.1, 0.1)``
    will use maximum 10% of the layouts total size.

    Children are resized based on ``size_hint``. Maximum and minimum
    ``size_hint``s only take effect if a ``size_hint`` is given.

    The layouts ``size_hint_min`` is automatically
    updated based on the minimal required space by children, after layouting.

    The width of columns and height of rows are calculated based on the size hints of the children.
    The highest size_hint_min of a child in a column or row is used. If a child has no size_hint,
    the actual size is considered.

    Args:
        x: ``x`` coordinate of bottom left corner.
        y: ``y`` coordinate of bottom left corner.
        width: Width of the layout.
        height: Height of the layout.
        align_horizontal: Align children in orthogonal direction.
            Options include ``left``, ``center``, and ``right``.
        align_vertical: Align children in orthogonal direction. Options
            include ``top``, ``center``, and ``bottom``.
        children: Initial list of children. More can be added later.
        size_hint: A size hint for :py:class:`~arcade.gui.UILayout`, if
            the :py:class:`~arcade.gui.UIWidget` would like to grow.
        size_hint_max: Maximum width and height in pixels.
        horizontal_spacing: Space between columns.
        vertical_spacing: Space between rows.
        column_count: Number of columns in the grid. This can be changed
            later.
        row_count: Number of rows in the grid. This can be changed
            later.
    """

    def __init__(
        self,
        *,
        x=0,
        y=0,
        width=1,
        height=1,
        align_horizontal="center",
        align_vertical="center",
        children: Iterable[UIWidget] = tuple(),
        size_hint=(0, 0),
        size_hint_max=None,
        horizontal_spacing: int = 0,
        vertical_spacing: int = 0,
        column_count: int = 1,
        row_count: int = 1,
        **kwargs,
    ):
        super().__init__(
            x=x,
            y=y,
            width=width,
            height=height,
            children=children,
            size_hint=size_hint,
            size_hint_max=size_hint_max,
            **kwargs,
        )
        self._size_hint_requires_update = True
        self._horizontal_spacing = horizontal_spacing
        self._vertical_spacing = vertical_spacing

        self.column_count = column_count
        self.row_count = row_count

        self.align_horizontal = align_horizontal
        self.align_vertical = align_vertical

        bind(self, "_children", self._trigger_size_hint_update)
        bind(self, "_border_width", self._trigger_size_hint_update)

        bind(self, "_padding_left", self._trigger_size_hint_update)
        bind(self, "_padding_right", self._trigger_size_hint_update)
        bind(self, "_padding_top", self._trigger_size_hint_update)
        bind(self, "_padding_bottom", self._trigger_size_hint_update)

        # initially update size hints
        # TODO is this required?
        self._update_size_hints()

    def add(
        self,
        child: W,
        column: int = 0,
        row: int = 0,
        column_span: int = 1,
        row_span: int = 1,
        **kwargs,
    ) -> W:
        """Add a widget to the grid layout.

        Args:
            child: Specified widget to add as a child of the layout.
            column: Column index in which the widget is to be added.
                The first column at the left of the widget starts at 0.
            row: The row number in which the widget is to be added.
                The first row at the top of the layout is numbered 0.
            column_span: Number of columns the widget will stretch for.
            row_span: Number of rows the widget will stretch for.
        """
        # subscribe to child's changes, which might affect the own size hint
        bind(child, "_children", self._trigger_size_hint_update)
        bind(child, "rect", self._trigger_size_hint_update)
        bind(child, "size_hint", self._trigger_size_hint_update)
        bind(child, "size_hint_min", self._trigger_size_hint_update)
        bind(child, "size_hint_max", self._trigger_size_hint_update)

        return super().add(
            child,
            column=column,
            row=row,
            column_span=column_span,
            row_span=row_span,
            **kwargs,
        )

    def remove(self, child: UIWidget):
        """Remove a child from the layout."""
        # unsubscribe from child's changes
        unbind(child, "_children", self._trigger_size_hint_update)
        unbind(child, "rect", self._trigger_size_hint_update)
        unbind(child, "size_hint", self._trigger_size_hint_update)
        unbind(child, "size_hint_min", self._trigger_size_hint_update)
        unbind(child, "size_hint_max", self._trigger_size_hint_update)

        return super().remove(child)

    def _trigger_size_hint_update(self):
        self._size_hint_requires_update = True

    def prepare_layout(self):
        """Updates the size hints if required."""
        super().prepare_layout()

        if self._size_hint_requires_update:
            self._update_size_hints()

    def _update_size_hints(self):
        self._size_hint_requires_update = False

        if not self.children:
            self.size_hint_min = (0, 0)
            return

        # 0. generate list for all rows and columns
        columns = []
        for i in range(self.column_count):
            columns.append([])
        rows = []
        for i in range(self.row_count):
            rows.append([])

        for entry in self._children:
            col_num = entry.data["column"]
            row_num = entry.data["row"]
            col_span = entry.data["column_span"]
            row_span = entry.data["row_span"]

            # we put the entry in all columns and rows it spans
            for c in range(col_span):
                columns[col_num + c].append(entry)

                for r in range(row_span):
                    rows[row_num + r].append(entry)

        # 1.a per column, collect max of size_hint_min and max size_hint
        minimal_width_per_column = []
        for col in columns:
            min_width = 0
            max_sh = 0
            for entry in col:
                col_span = entry.data["column_span"]
                # if the cell spans multiple columns,
                # we need to reduce the minimal required width by the horizontal spacing
                consumed_space = self._horizontal_spacing if col_span > 1 else 0

                min_w, _ = UILayout.min_size_of(entry.child)
                min_width = max(min_width, min_w / col_span - consumed_space)

                shw, _ = entry.child.size_hint or (0, 0)
                max_sh = max(max_sh, shw) if shw else max_sh

            minimal_width_per_column.append(min_width)

        # 1.b per row, collect max of size_hint_min and max size_hint
        minimal_height_per_row = []
        for row in rows:
            min_height = 0
            max_sh = 0
            for entry in row:
                row_span = entry.data["row_span"]
                # if the cell spans multiple rows,
                # we need to reduce the minimal required height by the vertical spacing
                consumed_space = self._vertical_spacing if row_span > 1 else 0

                _, min_h = UILayout.min_size_of(entry.child)
                min_height = max(min_height, min_h / row_span - consumed_space)

                _, shh = entry.child.size_hint or (0, 0)
                max_sh = max(max_sh, shh) if shh else max_sh

            minimal_height_per_row.append(min_height)

        base_width = self._padding_left + self._padding_right + 2 * self._border_width
        base_height = self._padding_top + self._padding_bottom + 2 * self._border_width

        self.size_hint_min = (
            base_width
            + sum(minimal_width_per_column)
            + (self.column_count - 1) * self._horizontal_spacing,
            base_height
            + sum(minimal_height_per_row)
            + (self.row_count - 1) * self._vertical_spacing,
        )

    def do_layout(self):
        """Executes the layout algorithm.

        Children are placed in a grid layout based on the size hints.

        Algorithm
        ---------

        0. generate list for all rows and columns
        1. per column, collect max of size_hint_min and max size_hint (widths)
        2. per row, collect max of size_hint_min and max size_hint (heights)
        3.  use box layout algorithm to distribute space
        4.  place widgets in grid layout

        """

        # skip if no children
        if not self.children:
            return

        # 0. generate list for all rows and columns
        columns = []
        for i in range(self.column_count):
            columns.append([])
        rows = []
        for i in range(self.row_count):
            rows.append([])

        lookup: dict[tuple[int, int], _ChildEntry] = {}
        for entry in self._children:
            col_num = entry.data["column"]
            row_num = entry.data["row"]
            col_span = entry.data["column_span"]
            row_span = entry.data["row_span"]

            # we put the entry in all columns and rows it spans
            for c in range(col_span):
                columns[col_num + c].append(entry)

                for r in range(row_span):
                    rows[row_num + r].append(entry)

            lookup[(col_num, row_num)] = entry

        # 1.a per column, collect max of size_hint_min and max size_hint
        minimal_width_per_column = []
        max_size_hint_per_column = []
        for col in columns:
            min_width = 0
            max_sh = 0
            for entry in col:
                col_span = entry.data["column_span"]
                # if the cell spans multiple columns,
                # we need to reduce the minimal required width by the horizontal spacing
                consumed_space = self._horizontal_spacing if col_span > 1 else 0

                min_w, _ = UILayout.min_size_of(entry.child)
                min_width = max(min_width, min_w / col_span - consumed_space)

                shw, _ = entry.child.size_hint or (0, 0)
                max_sh = max(max_sh, shw) if shw else max_sh

            minimal_width_per_column.append(min_width)
            max_size_hint_per_column.append(max_sh)

        # 1.b per row, collect max of size_hint_min and max size_hint
        minimal_height_per_row = []
        max_size_hint_per_row = []
        for row in rows:
            min_height = 0
            max_sh = 0
            for entry in row:
                row_span = entry.data["row_span"]
                # if the cell spans multiple rows,
                # we need to reduce the minimal required height by the vertical spacing
                consumed_space = self._vertical_spacing if row_span > 1 else 0

                _, min_h = UILayout.min_size_of(entry.child)
                min_height = max(min_height, min_h / row_span - consumed_space)

                _, shh = entry.child.size_hint or (0, 0)
                max_sh = max(max_sh, shh) if shh else max_sh

            minimal_height_per_row.append(min_height)
            max_size_hint_per_row.append(max_sh)

        # 2. use box layout algorithm to distribute space
        column_constraints = [
            _C(minimal_width_per_column[i], None, max_size_hint_per_column[i])
            for i in range(self.column_count)
        ]
        column_sizes = _box_axis_algorithm(
            column_constraints,
            self.content_width - (self.column_count - 1) * self._horizontal_spacing,
        )

        row_constraints = [
            _C(minimal_height_per_row[i], None, max_size_hint_per_row[i])
            for i in range(self.row_count)
        ]
        row_sizes = _box_axis_algorithm(
            row_constraints, self.content_height - (self.row_count - 1) * self._vertical_spacing
        )

        # 3. place widgets in grid layout
        start_y = self.content_rect.top
        for row_num in range(self.row_count):
            start_x = self.content_rect.left
            for col_num in range(self.column_count):
                entry = lookup.get((col_num, row_num))
                if not entry:
                    # still need to update start_x
                    start_x += column_sizes[col_num] + self._horizontal_spacing
                    continue

                # TODO handle row_span and col_span
                child = entry.child
                new_rect = child.rect

                # combine size of cells this entry spans and add spacing
                column_span = entry.data["column_span"]
                cell_width: float = sum(column_sizes[col_num : col_num + column_span])
                cell_width += (column_span - 1) * self._horizontal_spacing

                row_span = entry.data["row_span"]
                cell_height: float = sum(row_sizes[row_num : row_num + row_span])
                cell_height += (row_span - 1) * self._vertical_spacing

                # apply calculated sizes, when size_hint is given
                shw, shh = child.size_hint or (None, None)
                shmn_w, shmn_h = child.size_hint_min or (None, None)
                shmx_w, shmx_h = child.size_hint_max or (None, None)

                new_width = child.width
                if shw is not None:
                    new_width = min(cell_width, shw * self.content_width)
                    new_width = max(new_width, shmn_w or 0)
                    if shmx_w is not None:
                        new_width = min(new_width, shmx_w)

                new_height = child.height
                if shh is not None:
                    new_height = min(cell_height, shh * self.content_height)
                    new_height = max(new_height, shmn_h or 0)
                    if shmx_h is not None:
                        new_height = min(new_height, shmx_h)

                new_rect = new_rect.resize(width=new_width, height=new_height)

                # align within cell
                center_y = start_y - (cell_height / 2)
                center_x = start_x + (cell_width / 2)

                if self.align_vertical == "top":
                    new_rect = new_rect.align_top(start_y)
                elif self.align_vertical == "bottom":
                    new_rect = new_rect.align_bottom(start_y - row_sizes[row_num])
                else:
                    new_rect = new_rect.align_y(center_y)

                if self.align_horizontal == "left":
                    new_rect = new_rect.align_left(start_x)
                elif self.align_horizontal == "right":
                    new_rect = new_rect.align_right(start_x + cell_width)
                else:
                    new_rect = new_rect.align_x(center_x)

                # update child rect
                child.rect = new_rect

                start_x += column_sizes[col_num] + self._horizontal_spacing
            start_y -= row_sizes[row_num] + self._vertical_spacing


@dataclass
class _C:
    """Constrain values for the box algorithm.

    size_hint and min values of None are resolved to 0.0.
    """

    min: float
    max: float | None
    hint: float
    _final_size: float = 0.0
    """The final size of the entry which will be returned by the algorithm"""

    @staticmethod
    def from_widget(widget: UIWidget, dimension: Literal["width", "height"]) -> _C:
        index = 0 if dimension == "width" else 1

        # get hint values from different formats None and (float|None, float|None)
        sh = widget.size_hint[index] if widget.size_hint else None
        sh_min = widget.size_hint_min[index] if widget.size_hint_min else None
        sh_max = widget.size_hint_max[index] if widget.size_hint_max else None

        # resolve min and max values if no size hint is given
        min_value = widget.size[index] if sh is None else sh_min
        max_value = widget.size[index] if sh is None else sh_max

        # clean up None values
        min_value = min_value or 0
        sh = sh or 0.0

        return _C(
            min=min_value,
            max=max_value,
            hint=sh,
        )

    @staticmethod
    def from_widget_width(widget: UIWidget) -> _C:
        return _C.from_widget(widget, "width")

    @staticmethod
    def from_widget_height(widget: UIWidget) -> _C:
        return _C.from_widget(widget, "height")


def _box_orthogonal_algorithm(constraints: list[_C], container_size: float) -> list[float]:
    """Calculate the 1 dimensional size of each entry based on the hint value and the available
    space in the container.

    This calculation is done for the orthogonal axis of the box layout, which only applies the size
    hint to the orthogonal axis.

    Args:
        constraints: List of constraints with hint, min and max values
        container_size: The total size of the container
    """
    # calculate the width of each entry based on the hint
    for c in constraints:
        size = container_size * c.hint
        c.max = container_size if c.max is None else c.max

        c._final_size = min(max(c.min, size), c.max)  # clamp width to min and max values

    return [c._final_size for c in constraints]


def _box_axis_algorithm(constraints: list[_C], container_size: float) -> list[float]:
    """
    The box algorithm calculates the 1 dimensional size of each entry based on the hint value and
    the available space in the container.

    Args:
        constraints: List of constraints with hint, min and max values
        container_size: The total size of the container

    Returns:
        List of tuples with the sizes of each element
    """

    # if there is no space, return the min value of each constraint
    # this will cause a overflow, so we give a warning
    if container_size <= 0:
        warnings.warn("Container size is 0, cannot calculate sizes for children.")
        return [c.min for c in constraints]

    # adjust hint value based on min and max values
    for c in constraints:
        c.hint = max(c.min / container_size, c.hint)
        if c.max is not None:
            c.hint = min(c.hint, c.max / container_size)

    # normalize hint - which will cover cases, where the sum of the hints is greater than 1.
    # children will get a relative size based on their hint
    total_hint = sum(c.hint for c in constraints)
    if total_hint > 1:
        for c in constraints:
            c.hint /= total_hint

        # adjust hint value based on min values, again
        for c in constraints:
            c.hint = max(c.min / container_size, c.hint)

    # check if the total hint is greater than 1, again (caused by reapplied min values)
    total_hint = sum(c.hint for c in constraints)
    if total_hint > 1:
        # calculate the total hint value of all adjustable constraints
        total_adjustable_hint = sum(c.hint - c.min / container_size for c in constraints)

        # check if we have any adjustable constraints
        if total_adjustable_hint > 0:
            # reduce hint values of adjustable constraints to fit the container size
            required_adjustment = total_hint - 1
            possible_adjustment = min(required_adjustment, total_adjustable_hint)
            for c in constraints:
                adjustable_size = c.hint - c.min / container_size
                c.hint -= possible_adjustment * (adjustable_size / total_adjustable_hint)

    # calculate the width of each entry based on the hint
    for c in constraints:
        c._final_size = container_size * c.hint

    # return the calculated sizes, round to avoid floating point errors
    return [round(c._final_size, 5) for c in constraints]



"""Rects all act the same, but take four of the possible eight attributes and calculate the rest."""

from __future__ import annotations

import math
from typing import Any, NamedTuple, TypedDict

from pyglet.math import Vec2

from arcade.types.numbers import AsFloat
from arcade.types.vector_like import AnchorPoint, Point2

RectParams = tuple[AsFloat, AsFloat, AsFloat, AsFloat]
IntRectParams = tuple[int, int, int, int]


class RectKwargs(TypedDict):
    """Annotates a plain :py:class:`dict` of :py:class:`Rect` arguments.

    This is only meaningful as a type annotation during type checking.
    For example, the :py:meth:`Rect.kwargs <arcade.types.Rect.kwargs>`
    property returns an ordinary will actually be a :py:class:`dict`
    of :py:class:`Rect` field names to :py:class:`float` values.

    To learn more, please see:

    * :py:class:`.Rect`
    * :py:class:`typing.TypedDict`

    """

    left: float
    right: float
    bottom: float
    top: float
    width: float
    height: float
    x: float
    y: float


class Rect(NamedTuple):
    """A rectangle, with several convenience properties and functions.

    This object is immutable by design. It provides no setters, and is a NamedTuple
    subclass.

    Attempts to implement all Rectangle functions used in the library, and to be a
    helpful tool for developers storing/manipulating rectangle and rectangle-like
    constructs.

    Rectangles cannot rotate by design, since this complicates their implementation
    a lot.

    You probably don't want to create one of these directly, and should instead use
    a helper method, like :py:func:`.LBWH`, :py:func:`.LRBT`, :py:func:`.XYWH`, or
    :py:func:`.Viewport`.

    You can also use :py:meth:`.from_kwargs` to create a Rect from keyword arguments.
    """

    #: The X position of the rectangle's left edge.
    left: float
    #: The X position of the rectangle's right edge.
    right: float
    #: The Y position of the rectangle's bottom edge.
    bottom: float
    #: The Y position of the rectangle's top edge.
    top: float
    #: The total width of the rectangle along the X axis.
    #: To get the rectangle's :py:attr:`.height` well, use
    #: :py:attr:`.size`
    width: float
    #: The total height of the rectangle along the Y axis.
    #: To get the rectangle's :py:attr:`.width` as well, use
    #: :py:attr:`.size`.
    height: float
    #: The center of the rectangle along the X axis. To get its
    #: center :py:attr:`.y` as well, use :py:attr:`.center`.
    x: float
    #: The center of the rectangle along the Y axis. To get its
    #: center :py:attr:`.x` as well, use :py:attr:`.center`.
    y: float

    @property
    def center_x(self) -> float:
        """Backwards-compatible alias for :py:attr:`.x`."""
        return self.x

    @property
    def center_y(self) -> float:
        """Backwards-compatible alias for :py:attr:`.y`."""
        return self.y

    @property
    def center(self) -> Vec2:
        """Returns a :py:class:`~pyglet.math.Vec2` representing the center of the rectangle."""
        return Vec2(self.x, self.y)

    @property
    def bottom_left(self) -> Vec2:
        """
        Returns a :py:class:`~pyglet.math.Vec2` representing the
        bottom-left of the rectangle.
        """
        return Vec2(self.left, self.bottom)

    @property
    def bottom_right(self) -> Vec2:
        """
        Returns a :py:class:`~pyglet.math.Vec2` representing the
        bottom-right of the rectangle.
        """
        return Vec2(self.right, self.bottom)

    @property
    def top_left(self) -> Vec2:
        """
        Returns a :py:class:`~pyglet.math.Vec2` representing the
        top-left of the rectangle.
        """
        return Vec2(self.left, self.top)

    @property
    def top_right(self) -> Vec2:
        """
        Returns a :py:class:`~pyglet.math.Vec2` representing the
        top-right of the rectangle.
        """
        return Vec2(self.right, self.top)

    @property
    def bottom_center(self) -> Vec2:
        """
        Returns a :py:class:`~pyglet.math.Vec2` representing the
        bottom-center of the rectangle.
        """
        return Vec2(self.x, self.bottom)

    @property
    def center_right(self) -> Vec2:
        """
        Returns a :py:class:`~pyglet.math.Vec2` representing the
        center-right of the rectangle.
        """
        return Vec2(self.right, self.y)

    @property
    def top_center(self) -> Vec2:
        """
        Returns a :py:class:`~pyglet.math.Vec2` representing the
        top-center of the rectangle.
        """
        return Vec2(self.x, self.top)

    @property
    def center_left(self) -> Vec2:
        """Returns a :py:class:`~pyglet.math.Vec2` representing the center-left of the rectangle."""
        return Vec2(self.left, self.y)

    @property
    def size(self) -> Vec2:
        """Returns a :py:class:`~pyglet.math.Vec2` representing the size of the rectangle."""
        return Vec2(self.width, self.height)

    @property
    def area(self) -> float:
        """The area of the rectangle in square pixels."""
        return self.width * self.height

    @property
    def aspect_ratio(self) -> float:
        """Returns the ratio between the width and the height."""
        return self.width / self.height

    def at_position(self, position: Point2) -> Rect:
        """Returns a new :py:class:`Rect` which is moved to put `position` at its center."""
        x, y = position
        return XYWH(x, y, self.width, self.height)

    def move(self, dx: AsFloat = 0.0, dy: AsFloat = 0.0) -> Rect:
        """
        Returns a new :py:class:`Rect` which is moved by `dx` in the
        x-direction and `dy` in the y-direction.
        """
        return XYWH(self.x + dx, self.y + dy, self.width, self.height)

    def resize(
        self,
        width: AsFloat | None = None,
        height: AsFloat | None = None,
        anchor: Vec2 = AnchorPoint.CENTER,
    ) -> Rect:
        """
        Returns a new :py:class:`Rect` at the current Rect's position,
        but with a new width and height, anchored at a point (default center.)
        """
        width = width if width is not None else self.width
        height = height if height is not None else self.height

        anchor_x = self.left + anchor.x * self.width
        anchor_y = self.bottom + anchor.y * self.height

        # ratio_x = width / (self.width or 1.0)
        # ratio_y = height / (self.height or 1.0)

        # adjusted_left = anchor_x - (self.left - anchor_x) * ratio_x
        # adjusted_bottom = anchor_y + (self.bottom - anchor_y) * ratio_y
        # return LBWH(adjusted_left, adjusted_bottom, width, height)

        return XYWH(anchor_x, anchor_y, width, height, anchor)

    def scale(self, new_scale: AsFloat, anchor: Vec2 = AnchorPoint.CENTER) -> Rect:
        """
        Returns a new :py:class:`Rect` scaled by a factor of ``new_scale``,
        anchored at a point (default center.)
        """
        anchor_x = self.left + anchor.x * self.width
        anchor_y = self.bottom + anchor.y * self.height

        adjusted_left = anchor_x + (self.left - anchor_x) * new_scale
        adjusted_right = anchor_x + (self.right - anchor_x) * new_scale
        adjusted_top = anchor_y + (self.top - anchor_y) * new_scale
        adjusted_bottom = anchor_y + (self.bottom - anchor_y) * new_scale

        return LRBT(adjusted_left, adjusted_right, adjusted_bottom, adjusted_top)

    def scale_axes(self, new_scale: Point2, anchor: Vec2 = AnchorPoint.CENTER) -> Rect:
        """
        Return a new :py:class:`Rect`
        scaled by a factor of `new_scale.x` in the width
        and `new_scale.y` in the height, anchored at a point (default center.)
        """
        anchor_x = self.left + anchor.x * self.width
        anchor_y = self.bottom + anchor.y * self.height

        nsx, nsy = new_scale
        adjusted_left = anchor_x + (self.left - anchor_x) * nsx
        adjusted_right = anchor_x + (self.right - anchor_x) * nsx
        adjusted_top = anchor_y + (self.top - anchor_y) * nsy
        adjusted_bottom = anchor_y + (self.bottom - anchor_y) * nsy

        return LRBT(adjusted_left, adjusted_right, adjusted_bottom, adjusted_top)

    def __mul__(self, scale: AsFloat) -> Rect:  # type: ignore[override]
        """Scale the Rect by ``scale`` relative to ``(0, 0)``."""
        return Rect(
            self.left * scale,
            self.right * scale,
            self.bottom * scale,
            self.top * scale,
            self.width * scale,
            self.height * scale,
            self.x * scale,
            self.y * scale,
        )

    def __truediv__(self, scale: AsFloat) -> Rect:
        """Scale the rectangle by 1/``scale`` relative to ``(0, 0)``."""
        return Rect(
            self.left / scale,
            self.right / scale,
            self.bottom / scale,
            self.top / scale,
            self.width / scale,
            self.height / scale,
            self.x / scale,
            self.y / scale,
        )

    def align_top(self, value: AsFloat) -> Rect:
        """Returns a new :py:class:`Rect`, which is aligned to the top at `value`."""
        return LBWH(self.left, value - self.height, self.width, self.height)

    def align_bottom(self, value: AsFloat) -> Rect:
        """Returns a new :py:class:`Rect`, which is aligned to the bottom at `value`."""
        return LBWH(self.left, value, self.width, self.height)

    def align_left(self, value: AsFloat) -> Rect:
        """Returns a new :py:class:`Rect`, which is aligned to the left at `value`."""
        return LBWH(value, self.bottom, self.width, self.height)

    def align_right(self, value: AsFloat) -> Rect:
        """Returns a new :py:class:`Rect`, which is aligned to the right at `value`."""
        return LBWH(value - self.width, self.bottom, self.width, self.height)

    def align_center(self, value: Point2) -> Rect:
        """Returns a new :py:class:`Rect`, which is aligned to the center x and y at `value`."""
        cx, cy = value
        return XYWH(cx, cy, self.width, self.height)

    def align_x(self, value: AsFloat) -> Rect:
        """Returns a new :py:class:`Rect`, which is aligned to the x at `value`."""
        return XYWH(value, self.y, self.width, self.height)

    def align_center_x(self, value: AsFloat) -> Rect:
        """Backwards-compatible alias for `Rect.x`."""
        return self.align_x(value)

    def align_y(self, value: AsFloat) -> Rect:
        """Get a new :py:class:`Rect`, which is aligned to the y at `value`."""
        return XYWH(self.x, value, self.width, self.height)

    def align_center_y(self, value: AsFloat) -> Rect:
        """Backwards-compatible alias for `Rect.x`."""
        return self.align_y(value)

    def min_size(
        self,
        width: AsFloat | None = None,
        height: AsFloat | None = None,
        anchor: Vec2 = AnchorPoint.CENTER,
    ) -> Rect:
        """
        Return a :py:class:`Rect` that is at least size `width` by `height`, positioned at
        the current position and anchored to a point (default center.)
        """
        width = max(width or 0.0, self.width)
        height = max(height or 0.0, self.height)
        return self.resize(width, height, anchor)

    def max_size(
        self,
        width: AsFloat | None = None,
        height: AsFloat | None = None,
        anchor: Vec2 = AnchorPoint.CENTER,
    ) -> Rect:
        """
        Return a :py:class:`Rect` that is at most size `width` by `height`, positioned at
        the current position and anchored to a point (default center.)
        """
        width = min(width or float("inf"), self.width)
        height = min(height or float("inf"), self.height)
        return self.resize(width, height, anchor)

    def clamp_height(
        self,
        min_height: AsFloat | None = None,
        max_height: AsFloat | None = None,
        anchor: Vec2 = AnchorPoint.CENTER,
    ) -> Rect:
        """
        Return a :py:class:`Rect` that has a height between `min_height` and `max_height`,
        positioned at the current position and anchored to a point (default center.)
        """
        height = min(max_height or float("inf"), max(min_height or 0.0, self.height))
        return self.resize(self.width, height, anchor)

    def clamp_width(
        self,
        min_width: AsFloat | None = None,
        max_width: AsFloat | None = None,
        anchor: Vec2 = AnchorPoint.CENTER,
    ) -> Rect:
        """Return a :py:class:`.Rect` constrained to the passed dimension.

        It will be created as follows:

        * Its :py:attr:`.width` will be between any provided ``min_width``
          and ``max_width``
        * It will be positioned at the current position using the passed
          ``anchor``

        Args:
            min_width:
                An optional minimum width.
            max_width:
                An optional maximum width.
            anchor:
                A :py:class:`~pyglet.math.Vec2` of the fractional
                percentage of the rectangle's total :py:attr:`.size` along
                both axes. It defaults to the center.
        """
        width = min(max_width or float("inf"), max(min_width or 0.0, self.width))
        return self.resize(width, self.height, anchor)

    def clamp_size(
        self,
        min_width: AsFloat | None = None,
        max_width: AsFloat | None = None,
        min_height: AsFloat | None = None,
        max_height: AsFloat | None = None,
        anchor: Vec2 = AnchorPoint.CENTER,
    ) -> Rect:
        """Get a new clamped-size rectangle at the same position and anchored at ``anchor_point``.

        This combines the effects of :py:meth:`clamp_width` and :py:meth:`clamp_height` into
        one call.
        """
        width = min(max_width or float("inf"), max(min_width or 0.0, self.width))
        height = min(max_height or float("inf"), max(min_height or 0.0, self.height))
        return self.resize(width, height, anchor)

    def union(self, other: Rect) -> Rect:
        """Get the smallest rectangle covering both this one and ``other``."""
        left = min(self.left, other.left)
        right = max(self.right, other.right)
        bottom = min(self.bottom, other.bottom)
        top = max(self.top, other.top)
        return LRBT(left, right, bottom, top)

    def __or__(self, other: Rect) -> Rect:
        """Shorthand for :py:meth:`rect.union(other) <union>`.

        Args:
            other: Another :py:class:`Rect` instance.
        """
        return self.union(other)

    def intersection(self, other: Rect) -> Rect | None:
        """Return a :py:class:`Rect` of the overlap if any exists.

        If the two :py:class:`Rect` instances do not intersect, this
        method will return ``None`` instead.

        Args:
            other: Another :py:class:`Rect` instance.
        """
        intersecting = self.overlaps(other)
        if not intersecting:
            return None
        left = max(self.left, other.left)
        right = min(self.right, other.right)
        bottom = max(self.bottom, other.bottom)
        top = min(self.top, other.top)
        return LRBT(left, right, bottom, top)

    def __and__(self, other: Rect) -> Rect | None:
        """Shorthand for :py:meth:`rect.intersection(other) <interesection>`.

        Args:
            other: Another :py:class:`Rect` instance.
        """
        return self.intersection(other)

    def overlaps(self, other: Rect) -> bool:
        """Returns ``True`` if `other` overlaps with ``self``.

        Args:
            other: Another :py:class:`Rect` instance.
        """

        return (other.width + self.width) / 2.0 > abs(self.x - other.x) and (
            other.height + self.height
        ) / 2.0 > abs(self.y - other.y)

    def point_in_rect(self, point: Point2) -> bool:
        """Returns ``True`` if ``point`` is inside this rectangle.

        Args:
            point: A tuple of :py:class:`int` or :py:class:`float` values.
        """
        px, py = point
        return (self.left <= px <= self.right) and (self.bottom <= py <= self.top)

    def point_in_bounds(self, point: Point2) -> bool:
        """Returns ``True`` if ``point`` is inside this rectangle excluding the boundaries.

        Args:
            point: A tuple of :py:class:`int` or :py:class:`float` values.
        """
        px, py = point
        return (self.left < px < self.right) and (self.bottom < py < self.top)

    def __contains__(self, point: Point2 | Any) -> bool:
        """Shorthand for :py:meth:`rect.point_in_rect(point) <point_in_rect>`.

        Args:
            point: A tuple of :py:class:`int` or :py:class:`float` values.
        """
        from arcade.utils import is_iterable

        if not is_iterable(point):
            return False

        return self.point_in_rect(point)

    def distance_from_bounds(self, point: Point2) -> float:
        """Returns the point's distance from the boundary of this rectangle."""
        px, py = point
        diff = Vec2(px - self.x, py - self.y)
        dx = abs(diff.x) - self.width / 2.0
        dy = abs(diff.y) - self.height / 2.0
        d = (max(dx, 0.0) ** 2 + max(dy, 0.0) ** 2) ** 0.5 + min(max(dx, dy), 0.0)
        return d

    def point_on_bounds(self, point: Point2, tolerance: float) -> bool:
        """Returns ``True`` if ``point`` is within ``tolerance`` of the bounds.

        The ``point``'s distance from the bounds is computed by through
        :py:meth:`distance_from_bounds`.

        Args:
            point: The point to check.
            tolerance: The maximum distance the point can be from the bounds.
        """
        return abs(self.distance_from_bounds(point)) < tolerance

    def position_to_uv(self, point: Point2) -> Vec2:
        """Convert a point to UV space values inside the rectangle.

        This is like a pair of ratios which measure how far the ``point``
        is from the rectangle's :py:meth:`bottom_left` up toward its
        :py:meth:`top_right` along each axis.

        .. warning:: This method does not clamp output!

                     Since ``point`` is absolute pixels, one or both axes
                     of the returned :py:class:`~pyglet.math.Vec2` can be:

                     * less than ``0.0``
                     * greater than ``1.0``


        Each axis of the return value measures how far into
        the rectangle's ``size`` the ``point`` is relative
        to the :py:meth:`bottom_left`:

        .. code-block:: python

           # consult the diagram below
           Vec2(
             (point.x - rect.left)   / rect.width,
             (point.y - rect.bottom) / rect.height
           )

        .. code-block::

              |------- rect.width ------|

                     The rectangle      (rect.top_right)
              +-------------------------T    ---
              |                         |     |
           - - - - - - - P  (Point x, y)|     |
           |  |                         | rect.height
           |  |          |              |     |
           y  |                         |     |
           |  B----------|--------------+    ---
           |  (rect.bottom_right)
           |
           O----- x -----|

        Args:
            point: A point relative to the rectangle's :py:meth:`bottom_left` corner.
        """
        x, y = point
        return Vec2(
            (x - self.left) / self.width,
            (y - self.bottom) / self.height,
        )

    def uv_to_position(self, uv: Point2) -> Vec2:
        """Convert a point in UV-space to a point within the rectangle.

        The ``uv`` is a pair of ratios which describe how far a point
        extends across the rectangle's :py:attr:`width` and
        :py:attr:`height` from the :py:attr:`bottom_left` toward its
        :py:attr:`top_right`.

        .. warning:: This method does not clamp output!

                     Since one or both of ``uv``'s components can be
                     less than ``0.0`` or greater than ``1.0``, the
                     returned point can fall outside the rectangle.

        The following can be used as arguments to this function:

        #. Values in :py:class:`~arcade.types.AnchorPoint`
        #. Returned values from :py:meth:`position_to_uv`
        #. Rescaled input data from controllers

        Args:
            uv:
                A pair of ratio values describing how far a
                a point falls from a rectangle's :py:attr:`bottom_left`
                toward its :py:attr:`top_right`.
        """
        x, y = uv
        return Vec2(self.left + x * self.width, self.bottom + y * self.height)

    def get_relative_to_anchor(self, point: Point2, anchor: Vec2 = AnchorPoint.CENTER) -> Vec2:
        """Convert a point to a relative offset from the anchor point.

        Args:
            point:
                The point to make relative.
            anchor:
                The anchor point to make the point relative to.
        """
        x, y = point
        rx = x - (self.left + (self.width * anchor.x))
        ry = y - (self.bottom + (self.height * anchor.y))
        return Vec2(rx, ry)

    def to_points(self) -> tuple[Vec2, Vec2, Vec2, Vec2]:
        """Return a new :py:class:`tuple` of this rectangle's corner points.

        The points will be ordered as follows:

        #. :py:meth:`bottom_left`
        #. :py:meth:`top_left`
        #. :py:meth:`top_right`
        #. :py:meth:`bottom_right`

        """
        left = self.left
        bottom = self.bottom
        right = self.right
        top = self.top
        return (Vec2(left, bottom), Vec2(left, top), Vec2(right, top), Vec2(right, bottom))

    @property
    def lbwh(self) -> RectParams:
        """Provides a tuple in the format of (left, bottom, width, height)."""
        return (self.left, self.bottom, self.width, self.height)

    @property
    def lrbt(self) -> RectParams:
        """Provides a tuple in the format of (left, right, bottom, top)."""
        return (self.left, self.right, self.bottom, self.top)

    @property
    def xywh(self) -> RectParams:
        """Provides a tuple in the format of (x, y, width, height)."""
        return (self.x, self.y, self.width, self.height)

    @property
    def xyrr(self) -> RectParams:
        """Provides a tuple in the format of (x, y, width / 2, height / 2)."""
        return (self.x, self.y, self.width / 2, self.height / 2)

    @property
    def lbwh_int(self) -> IntRectParams:
        """Provides a tuple in the format of (left, bottom, width, height), casted to ints."""
        return (int(self.left), int(self.bottom), int(self.width), int(self.height))

    @property
    def lrbt_int(self) -> IntRectParams:
        """Provides a tuple in the format of (left, right, bottom, top), casted to ints."""
        return (int(self.left), int(self.right), int(self.bottom), int(self.top))

    @property
    def xywh_int(self) -> IntRectParams:
        """Provides a tuple in the format of (x, y, width, height), casted to ints."""
        return (int(self.x), int(self.y), int(self.width), int(self.height))

    @property
    def xyrr_int(self) -> RectParams:
        """Provides a tuple in the format of (x, y, width / 2, height / 2), casted to ints."""
        return (int(self.x), int(self.y), int(self.width) / 2, int(self.height) / 2)

    @classmethod
    def from_kwargs(cls, **kwargs: AsFloat) -> Rect:
        """Creates a new Rect from keyword arguments. Throws ValueError if not enough are provided.

        Expected forms are:

        * LRBT (providing ``left``, ``right``, ``bottom``, and ``top``)
        * LBWH (providing ``left``, ``bottom``, ``width``, and ``height``)
        * XYWH (providing ``x``, ``y``, ``width``, and ``height``)
        """
        # Perform iteration only once and store it as a set literal
        specified: set[str] = {k for k, v in kwargs.items() if v is not None}
        have_lb = "left" in specified and "bottom" in specified

        # LRBT
        if have_lb and "top" in specified and "right" in specified:
            return LRBT(kwargs["left"], kwargs["right"], kwargs["bottom"], kwargs["top"])  # type: ignore

        # LBWH
        have_wh = "width" in specified and "height" in specified
        if have_wh and have_lb:
            return LBWH(kwargs["left"], kwargs["bottom"], kwargs["width"], kwargs["height"])  # type: ignore

        # XYWH
        if have_wh and "x" in specified and "y" in specified:
            return XYWH(kwargs["x"], kwargs["y"], kwargs["width"], kwargs["height"])  # type: ignore

        raise ValueError("Not enough attributes defined for a valid rectangle!")

    @property
    def kwargs(self) -> RectKwargs:
        """Get this rectangle as a :py:class:`dict` of field names to values.

        .. _tomli: https://github.com/hukkin/tomli

        Many data formats have corresponding Python modules with write
        support. Such modules often one or more functions which convert
        a passed :py:class:`dict` to a :py:class:`str` or write the result
        of such a conversion to a file.

        For example, the built-in :py:mod:`json` module offers the
        following functions on all Python versions currently supported
        by Arcade:

        .. list-table::
           :header-rows: 1

           * - Function
             - Summary
             - Useful For

           * - :py:func:`json.dump`
             - Write a :py:class:`dict` to a file
             - Saving game progress or edited levels

           * - :py:func:`json.dumps`
             - Get a :py:class:`dict` as a :py:class:`str` of JSON
             - Calls to a Web API

        .. note::

           The return value is an ordinary :py:class:`dict`.

           Although the return type is annotated as a
           :py:class:`.RectKwargs`, it is only meaningful when type
           checking. See :py:class:`typing.TypedDict` to learn more.

        """
        return {
            "left": self.left,
            "right": self.right,
            "bottom": self.bottom,
            "top": self.top,
            "x": self.x,
            "y": self.y,
            "width": self.width,
            "height": self.height,
        }

    # Since __repr__ is handled automatically by NamedTuple, we focus on
    # human-readable spot-check values for __str__ instead.
    def __str__(self) -> str:
        return (
            f"<{self.__class__.__name__} LRBT({self.left}, {self.right}, {self.bottom}, {self.top})"
            f" XYWH({self.x}, {self.y}, {self.width}, {self.height})>"
        )

    def __bool__(self) -> bool:
        """Returns True if area is not 0, else False."""
        return self.width != 0 and self.height != 0

    def __round__(self, n: int) -> Rect:
        """Rounds the left, right, bottom, and top to `n` decimals."""
        return LRBT(
            round(self.left, n), round(self.right, n), round(self.bottom, n), round(self.top, n)
        )

    def __floor__(self) -> Rect:
        """Floors the left, right, bottom, and top."""
        return LRBT(
            math.floor(self.left),
            math.floor(self.right),
            math.floor(self.bottom),
            math.floor(self.top),
        )

    def __ceil__(self) -> Rect:
        """Floors the left, right, bottom, and top."""
        return LRBT(
            math.ceil(self.left), math.ceil(self.right), math.ceil(self.bottom), math.ceil(self.top)
        )


# Shorthand creation helpers


def LRBT(left: AsFloat, right: AsFloat, bottom: AsFloat, top: AsFloat) -> Rect:
    """Creates a new :py:class:`.Rect` from left, right, bottom, and top parameters."""
    width = right - left
    height = top - bottom
    x = left + (width / 2)
    y = bottom + (height / 2)
    return Rect(left, right, bottom, top, width, height, x, y)


def LBWH(left: AsFloat, bottom: AsFloat, width: AsFloat, height: AsFloat) -> Rect:
    """Creates a new :py:class:`.Rect` from left, bottom, width, and height parameters."""
    right = left + width
    top = bottom + height
    x = left + (width / 2)
    y = bottom + (height / 2)
    return Rect(left, right, bottom, top, width, height, x, y)


def XYWH(
    x: AsFloat, y: AsFloat, width: AsFloat, height: AsFloat, anchor: Vec2 = AnchorPoint.CENTER
) -> Rect:
    """
    Creates a new :py:class:`.Rect` from x, y, width, and height parameters,
    anchored at a relative point (default center).
    """
    left = x - anchor.x * width
    right = left + width
    bottom = y - anchor.y * height
    top = bottom + height
    c_x = left + width / 2.0
    c_y = bottom + height / 2.0
    return Rect(left, right, bottom, top, width, height, c_x, c_y)


def XYRR(x: AsFloat, y: AsFloat, half_width: AsFloat, half_height: AsFloat) -> Rect:
    """
    Creates a new :py:class:`.Rect` from center x, center y, half width, and half height parameters.
    This is mainly used by OpenGL.
    """
    left = x - half_width
    right = x + half_width
    bottom = y - half_height
    top = y + half_height
    return Rect(left, right, bottom, top, half_width * 2, half_height * 2, x, y)


def Viewport(left: int, bottom: int, width: int, height: int) -> Rect:
    """
    Creates a new :py:class:`.Rect` from left, bottom, width, and height parameters,
    restricted to integers.
    """
    right = left + width
    top = bottom + height
    x = left + int(width / 2)
    y = bottom + int(height / 2)
    return Rect(left, right, bottom, top, width, height, x, y)


__all__ = [
    "IntRectParams",
    "RectParams",
    "RectKwargs",
    "Rect",
    "AnchorPoint",
    "LBWH",
    "LRBT",
    "XYWH",
    "XYRR",
    "Viewport",
]



import array

from arcade.types import Color, Point2List, RGBOrA255
from arcade.types.rect import XYWH
from arcade.window_commands import get_window

from .rect import draw_rect_filled


def draw_point(x: float, y: float, color: RGBOrA255, size: float = 1.0) -> None:
    """Draw a 2D point at ``(x, y)`` as a square ``size`` pixels wide.

    The square will be centered on ``(x, y)`` with the given ``color``
    as its fill value.

    To draw more rounded shapes, please see:

    * :py:func:`arcade.draw.circle.draw_circle_filled`
    * :py:func:`pyglet.shapes.Circle`

    Args:
        x:
            The center of the square along the x axis.
        y:
            The center of the square along the y axis.
        color:
            The fill color of the square as an RGBA :py:class:`tuple`,
            RGB py:class:`tuple`, or a :py:class:`.Color` instance.
        size:
             The width and height of the square in pixels.
    """
    draw_rect_filled(XYWH(x, y, size, size), color)


def draw_points(point_list: Point2List, color: RGBOrA255, size: float = 1.0) -> None:
    """Draw 2D points as squares ``size`` pixels  wide.

    Each point in ``point_list`` will be drawn centered on its x and y
    value with the same ``color`` and square ``size`` in pixels.

    To draw more rounded shapes, please see:

    * :py:func:`arcade.draw.circle.draw_circle_filled`
    * :py:func:`pyglet.shapes.Circle`

    Args:
        point_list:
            A :py:class:`list` or :py:class:`tuple` of 2D points.
        color:
            The fill color for the points as an RGBA :py:class:`tuple`,
            RGB :py:class:`tuple`, or :py:class:`.Color` instance.
        size:
            The width and height of each point's square in pixels.
    """
    # Fails immediately if we don't have a window or context
    window = get_window()
    ctx = window.ctx
    program = ctx.shape_rectangle_filled_unbuffered_program  # type: ignore
    geometry = ctx.shape_rectangle_filled_unbuffered_geometry
    buffer = ctx.shape_rectangle_filled_unbuffered_buffer  # type: ignore

    # Validate & normalize to a pass the shader an RGBA float uniform
    color_normalized = Color.from_iterable(color).normalized

    # Get # of points and translate Python tuples to a C-style array
    num_points = len(point_list)
    point_array = array.array("f", (v for point in point_list for v in point))

    # Resize buffer
    data_size = num_points * 8
    # if data_size > buffer.size:
    buffer.orphan(size=data_size)

    ctx.enable(ctx.BLEND)

    # Pass data to shader
    program["color"] = color_normalized
    program["shape"] = size, size, 0
    buffer.write(data=point_array)

    # Only render the # of points we have complete data for
    geometry.render(program, mode=ctx.POINTS, vertices=data_size // 8)

    ctx.disable(ctx.BLEND)



"""
Simple shader based noise example

The code creates squares of random colors all around the Window.
Uses code injection to use the built-in random.glsl file. This allows for easy access
to a high quality hash function and methods of using 1, 2, 3, or 4 floats as seeds.
The function generates a value between 0 and 1.

It is very important when using the x and y coordinates to try use both in all instances
If you only use one you will get very obvious artifacts all along the axis that not used.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.random_shader_simple
"""

from time import time
from array import array
from random import random

from arcade import Window
from arcade.gl import BufferDescription


def generate_points(count: int):
    for _ in range(count):
        yield random() * 2 - 1
        yield random() * 2 - 1


class App(Window):

    def __init__(self):
        super().__init__()
        # the number of points we want showing up
        self._point_count = 128
        # so that the colours change every run we store the time at run to use as a seed
        self._time_seed = time()
        self._program = self.ctx.program(
            # We're passing the shader source to ctx.shader_inc() so that we can use
            # the #include directive. This is not needed when using load_program() as
            # it will automatically look for the file.
            vertex_shader=self.ctx.shader_inc(
                """
                #version 330

                in vec2 in_pos;

                // This in the index of this specific vertex.
                // Flat just specifies that the value shouldn't be interpolated between vertices.
                flat out int vert_id;
                out vec2 vert_pos;

                void main(){
                    gl_Position = vec4(in_pos, 0.0, 1.0);

                    vert_id = gl_VertexID;
                    vert_pos = in_pos;
                }
                """
            ),
            fragment_shader=self.ctx.shader_inc("""
            #version 330

            #include :resources:/shaders/lib/random.glsl

            // Predefining the function which will be over written with the code injection.
            // the random function takes in 1, 2, 3, or 4 floats and returns a new float
            // between 0 and 1
            float random(vec2 v);
            float random(vec3 v);
            float random(vec4 v);

            // Both of these are used as seed values to make sure each call is unique
            uniform float time_seed;
            flat in int vert_id;

            // so each pixel in each vertex has the same color we pass in the position
            in vec2 vert_pos;

            out vec4 frag_colour;

            void main() {
                float red = random(vec4(vert_pos.x, vert_pos.y, time_seed, vert_id));
                float green = random(vec4(vert_id, time_seed, vert_pos.y, vert_pos.x));
                float blue = random(vec4(vert_pos.y, vert_pos.x, time_seed, vert_id));
                frag_colour = vec4(red, green, blue, 1.0);
            }
            """),
        )
        self._program["time_seed"] = self._time_seed

        self._geometry = self.ctx.geometry(
            [
                BufferDescription(
                    self.ctx.buffer(data=array("f", generate_points(self._point_count))),
                    "2f",
                    ["in_pos"],
                )
            ],
            mode=self.ctx.POINTS,
        )

    def on_draw(self):
        self.clear()
        self.ctx.point_size = 16.0
        self._geometry.render(self._program)


App().run()
